'''
@copyright: 2022 - Symas Corporation
'''

from rbac.util import Config

DELIMITER = Config.get('schema')['raw_delimiter']

class Constraint:    
    "Fortress Constraint"
    
    def __init__(
            self,
            raw=None,
            name=None,
            timeout=None,
            begin_time=None,
            end_time=None,
            begin_date=None,
            end_date=None,
            begin_lock_date=None,
            end_lock_date=None,
            day_mask=None 
            ):
    
        self.raw = raw
        # If ldap entity doesn't have constraint, this happens:
        if self.raw is not None and not self.raw:
            pass
        # If ldap entity has constraint:        
        elif self.raw is not None:
            entity_constraint = self.raw.split(DELIMITER)
            entity_constraint = [ val.strip() for val in entity_constraint ]
            if entity_constraint[0] is not None:
                self.name=entity_constraint[0]
            if entity_constraint[1] is not None:
                self.timeout=int(entity_constraint[1])        
            if entity_constraint[2] is not None:
                self.begin_time=entity_constraint[2]
            if entity_constraint[3] is not None:
                self.end_time=entity_constraint[3]
            if entity_constraint[4] is not None:
                self.begin_date=entity_constraint[4]
            if entity_constraint[5] is not None:
                self.end_date=entity_constraint[5]
            if entity_constraint[6] is not None:
                self.begin_lock_date=entity_constraint[6]
            if entity_constraint[7] is not None:
                self.end_lock_date=entity_constraint[7]
            if entity_constraint[8] is not None:
                self.day_mask=entity_constraint[8]
        # if going the other way, from caller to ldap, this will occur:                            
        else:
            self.name = name
            self.timeout = timeout
            self.begin_time = begin_time
            self.end_time = end_time
            self.begin_date = begin_date
            self.end_date = end_date
            self.begin_lock_date = begin_lock_date
            self.end_lock_date = end_lock_date
            self.day_mask = day_mask
            
    def get_raw(self):
        # Raw format: name$timeout$begin_time$end_time$begin_date$end_date$begin_lock_date$end_lock_date$day_mask
        # Example:
        # oamT12SSD1$30$0000$0000$20090101$21000101$20500101$20500115$1234567
        raw = self.name + DELIMITER
        if self.timeout is not None:
            raw += str(self.timeout)
        else:
            raw += '0'
        raw += DELIMITER
        if self.begin_time is not None:
            raw += self.begin_time
        raw += DELIMITER
        if self.end_time is not None:
            raw += self.end_time
        raw += DELIMITER
        if self.begin_date is not None:
            raw += self.begin_date
        raw += DELIMITER
        if self.end_date is not None:
            raw += self.end_date
        raw += DELIMITER
        if self.begin_lock_date is not None:
            raw += self.begin_lock_date
        raw += DELIMITER
        if self.end_lock_date is not None:
            raw += self.end_lock_date
        raw += DELIMITER
        if self.day_mask is not None:
            raw += self.day_mask     
        return raw       



'''
@copyright: 2022 - Symas Corporation
'''

from datetime import datetime


class CurrentDateTime:

    def __init__(
        self,
        ):
        
        now_=datetime.now()
        self.minute = str(now_.minute).zfill(2)
        self.hour = str(now_.hour).zfill(2)            
        self.month = str(now_.month).zfill(2)
        self.day = str(now_.day).zfill(2)
        self.year = str(now_.year).zfill(4)
        self.date = self.year + self.month + self.day
        self.time = self.hour + self.minute
        self.day_of_week = str(now_.weekday()+1)
        self.seconds = now_.timestamp()


'''
@copyright: 2022 - Symas Corporation
'''

import uuid
import ldap
from ldap import MOD_REPLACE, MOD_ADD, MOD_DELETE
from ldap.cidict import cidict as CIDict
from ldap.dn import str2dn
from rbac.model import Role, Constraint
from rbac.ldap import ldaphelper, NotFound, NotUnique
from rbac.ldap import userdao
from rbac.ldap.ldaphelper import add_to_modlist, mods_to_modlist
from rbac.util import global_ids
from rbac.util.fortress_error import RbacError

def read (entity):
    roleList = search(entity)
    if roleList is None or len(roleList) == 0:
        raise NotFound(msg="Role Read not found, name=" + entity.name, id=global_ids.ROLE_NOT_FOUND)    
    elif len(roleList) > 1:
        raise NotUnique(msg="Role Read not unique, name=" + entity.name, id=global_ids.ROLE_READ_FAILED)
    else:
        return roleList[0]


def search (entity):
    conn = None            
    roleList = []
    search_filter = '(&(objectClass=' + ROLE_OC_NAME + ')'
    search_filter += '(' + ROLE_NAME + '=' + entity.name + '))'
    try:
        conn = ldaphelper.open()
        entries = conn.search_s(__CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=SEARCH_ATTRS)
        for dn, attrs in entries:
            roleList.append(__unload(dn, attrs))
    except Exception as e:
        raise RbacError(msg='Role search error=' + str(e), id=global_ids.ROLE_SEARCH_FAILED)
    finally:
        if conn:        
            ldaphelper.close(conn)
    return roleList


def __unload(dn, attrs):
    entity = Role()
    entity.dn = dn

    attrs = CIDict(attrs)

    entity.internal_id = ldaphelper.get_attr_val(attrs.get(global_ids.INTERNAL_ID, []))
    entity.name = ldaphelper.get_attr_val(attrs.get(ROLE_NAME, []))
    entity.description = ldaphelper.get_one_attr_val(attrs.get(global_ids.DESC, []))
    # Get the multi-occurring attrs:
    entity.props = ldaphelper.get_list(attrs.get(global_ids.PROPS, []))
    entity.members = ldaphelper.get_list(attrs.get(MEMBER, []))
    # unload raw constraint:
    entity.constraint = Constraint(ldaphelper.get_attr_val(attrs.get(global_ids.CONSTRAINT, [])))
    return entity


def create ( entity ):
    try:
        attrs = {}
        attrs.update( {'objectClass': ROLE_OCS} )
        attrs.update( {global_ids.CN : entity.name} )
        attrs.update( {ROLE_NAME : entity.name} )
        # generate random id:
        entity.internal_id = str(uuid.uuid4())
        attrs.update( {global_ids.INTERNAL_ID : entity.internal_id} )        

        # string:
        if entity.description:        
            attrs.update( {global_ids.DESC : entity.description} )
        # list of strings
        if entity.props is not None and len(entity.props) > 0 :        
            attrs.update( {global_ids.PROPS : entity.props} )
        # list of comma delimited strings:
        if entity.constraint is None:
            entity.constraint = Constraint(name=entity.name)
        #if entity.constraint is not None:        
        attrs.update( {global_ids.CONSTRAINT : entity.constraint.get_raw()} )
            
        conn = ldaphelper.open()        
        conn.add_s(__get_dn(entity), add_to_modlist(attrs))
    except ldap.ALREADY_EXISTS:
        raise NotUnique(msg='Role create failed, already exists:' + entity.uid, id=global_ids.ROLE_ADD_FAILED)
    except ldap.LDAPError as e:
        raise RbacError(msg='Role create failed result=' + str(e), id=global_ids.ROLE_ADD_FAILED)
    except Exception as e:
        raise RbacError(msg='Role create error=' + str(e), id=global_ids.ROLE_ADD_FAILED)
    return entity


def update ( entity ):
    try:
        attrs = {}
        if entity.description:        
            attrs.update( {global_ids.DESC : [(MOD_REPLACE, [entity.description])]} )
        # list of srings:
        if entity.props is not None and len(entity.props) > 0 :        
            attrs.update( {global_ids.PROPS : [(MOD_REPLACE, [entity.props])]} )
        # list of comma delimited strings:
        if entity.constraint is not None :        
            attrs.update( {global_ids.CONSTRAINT : [(MOD_REPLACE, [entity.constraint.get_raw()])]} )
        if len(attrs) > 0:
            conn = ldaphelper.open()        
            conn.modify_s(__get_dn(entity), mods_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='Role update failed, not found:' + entity.name, id=global_ids.ROLE_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError('Role update failed result=' + str(e), global_ids.ROLE_UPDATE_FAILED)
    except Exception as e:
        raise RbacError(msg='Role update error=' + str(e), id=global_ids.ROLE_UPDATE_FAILED)
    return entity


def delete ( entity ):
    try:
        conn = ldaphelper.open()        
        conn.delete_s(__get_dn(entity))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='Role delete not found:' + entity.name, id=global_ids.ROLE_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='Role delete failed result=' + str(e), id=global_ids.ROLE_DELETE_FAILED)
    except Exception as e:
        raise RbacError(msg='Role delete error=' + str(e), id=global_ids.ROLE_DELETE_FAILED)
    return entity


def add_member ( entity, uid ):    
    try:        
        attrs = {}
        if uid:
            user_dn = __get_user_dn(uid)            
            attrs.update( {MEMBER : [(MOD_ADD, user_dn)]} )
            conn = ldaphelper.open()
            conn.modify_s(__get_dn(entity), mods_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='Add member failed, not found, role=' +  entity.name + ', member dn=' + user_dn, id=global_ids.ROLE_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='Add member failed result=' + str(e), id=global_ids.ROLE_USER_ASSIGN_FAILED)
    except Exception as e:
        raise RbacError(msg='Add member error=' + str(e), id=global_ids.ROLE_USER_ASSIGN_FAILED)
    return entity


def remove_member ( entity, uid ):    
    try:        
        attrs = {}
        if uid:
            user_dn = __get_user_dn(uid)
            attrs.update( {MEMBER : [(MOD_DELETE, user_dn)]} )
            conn = ldaphelper.open()
            conn.modify_s(__get_dn(entity), mods_to_modlist(attrs))
    except ldap.NO_SUCH_ATTRIBUTE:
        raise RbacError(msg='Remove member failed, not assigned, role=' +  entity.name + ', member dn=' + user_dn, id=global_ids.URLE_ASSIGN_NOT_EXIST)
    except ldap.LDAPError as e:
        raise RbacError(msg='Remove member failed result=' + str(e), id=global_ids.ROLE_REMOVE_OCCUPANT_FAILED)
    except Exception as e:
        raise RbacError(msg='Remove member error=' + str(e), id=global_ids.ROLE_REMOVE_OCCUPANT_FAILED)
    return entity


def get_members (entity):
    conn = None            
    uList = []
    search_filter = '(&(objectClass=' + ROLE_OC_NAME + ')'
    search_filter += '(' + ROLE_NAME + '=' + entity.name + '))'
    try:
        conn = ldaphelper.open()
        # TODO: use sizelimit=1
        entries = conn.search_s(__CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=SEARCH_ATTRS)

        if not entries:
            raise NotFound(msg="Role not found, name=" + entity.name, id=global_ids.ROLE_NOT_FOUND)
        elif len(entries) > 1:
            raise NotUnique(msg="Role not unique, name=" + entity.name, id=global_ids.ROLE_SEARCH_FAILED)

        dn, attrs = entries[0]

        member_dns = ldaphelper.get_list(attrs.get(MEMBER, []))
        uList = __convert_list(member_dns)
    except Exception as e: # FIXME: change to LDAPError
        raise RbacError(msg='Get members search error=' + str(e), id=global_ids.ROLE_OCCUPANT_SEARCH_FAILED)
    finally:
        if conn:        
            ldaphelper.close(conn)
    return uList


def get_members_constraint (entity):
    conn = None            
    mList = []
    search_filter = '(&(objectClass=' + ROLE_OC_NAME + ')'
    search_filter += '(' + ROLE_NAME + '=' + entity.name + '))'
    try:
        conn = ldaphelper.open()
        # TODO: use sizelimit=1
        entries = conn.search_s(__CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=[MEMBER, global_ids.CONSTRAINT])

        if not entries:
            raise NotFound(msg="Role not found, name=" + entity.name, id=global_ids.ROLE_NOT_FOUND)
        elif len(entries) > 1:
            raise NotUnique(msg="Role not unique, name=" + entity.name, id=global_ids.ROLE_SEARCH_FAILED)

        dn, attrs = entries[0]

        member_dns = ldaphelper.get_list(attrs.get(MEMBER, []))
        constraint = Constraint(ldaphelper.get_attr_val(attrs.get(global_ids.CONSTRAINT, [])))
        mList = __convert_list(member_dns)
    except Exception as e: # FIXME: change to LDAPError
        raise RbacError(msg='Get members search error=' + str(e), id=global_ids.ROLE_OCCUPANT_SEARCH_FAILED)
    finally:
        if conn:        
            ldaphelper.close(conn)
    return [mList, constraint]


def __get_dn(entity):
    return global_ids.CN + '=' + entity.name + "," + __CONTAINER_DN


def __get_user_dn(uid):
    # TODO: find a better way to do this:
    return global_ids.UID + '=' + uid + ',' + userdao.CONTAINER_DN


def __convert_list(list_dns):
    return [__get_user_id(dn) for dn in list_dns]


def __get_user_id(user_dn):
    dn = str2dn(user_dn)
    for attr, value, _ in dn[0]:
        if attr.lower() == "uid":
            return value
    

ROLE_OC_NAME = 'ftRls'
ROLE_OCS = [ROLE_OC_NAME, global_ids.PROP_OC_NAME]
ROLE_NAME = 'ftRoleName'
MEMBER = 'roleOccupant'

SEARCH_ATTRS = [
    global_ids.INTERNAL_ID, ROLE_NAME, global_ids.CONSTRAINT, global_ids.PROPS, global_ids.DESC, MEMBER
     ]

__CONTAINER_DN = ldaphelper.get_container_dn(global_ids.ROLE_OU)



'''
@copyright: 2022 - Symas Corporation
'''

from . import Validator
from .global_ids import ACTV_FAILED_DATE, SUCCESS

class Date(Validator):
        
    def validate(self, constraint, now, session=None):      
        if not constraint.begin_date or constraint.begin_date is None or not constraint.end_date or constraint.end_date is None:
            return SUCCESS        
        elif constraint.begin_date == 'none' or constraint.end_date == 'none':
            return SUCCESS        
        elif constraint.begin_date <= now.date <= constraint.end_date:
            return SUCCESS
        else:            
            return ACTV_FAILED_DATE


'''
@copyright: 2022 - Symas Corporation
'''

from rbac.util.fortress_error import RbacError
from rbac.util import global_ids


def validate_user(user):
    if user is None:
        raise RbacError (msg='User is None', id=global_ids.USER_NULL)
    elif not user.uid:
        raise RbacError (msg='User uid is None', id=global_ids.USER_ID_NULL)


def validate_role(role):
    if role is None:
        raise RbacError (msg='Role is None', id=global_ids.ROLE_NULL)
    elif not role.name:
        raise RbacError (msg='Role name is None', id=global_ids.ROLE_NM_NULL)


def validate_perm_obj(perm):
    if perm is None:
        raise RbacError (msg='Perm object is None', id=global_ids.PERM_OBJECT_NULL)
    elif not perm.obj_name:
        raise RbacError (msg='Perm object name is None', id=global_ids.PERM_OBJECT_NM_NULL)

    
def validate_perm(perm):
    if perm is None:
        raise RbacError (msg='Permission is None', id=global_ids.PERM_OPERATION_NULL)
    elif not perm.obj_name:
        raise RbacError (msg='Permission object name is None', id=global_ids.PERM_OBJECT_NM_NULL)
    elif not perm.op_name:
        raise RbacError (msg='Permission operation name is None', id=global_ids.PERM_OPERATION_NM_NULL)



'''
@copyright: 2022 - Symas Corporation
'''

from rbac.ldap import permdao, userdao, roledao
from rbac.util import global_ids, utils
from rbac.util.fortress_error import RbacError
from rbac.model import Perm, User, Role
from rbac.util import logger

def add_user(user):
    """
    This command creates a new RBAC user. The command is valid only if the new user is not already a member of the USERS data set. 
    The USER data set is updated. The new user does not own any session at the time of its creation.
    
    required parameters:
    user.uid - maps to INetOrgPerson uid
         
    optional parameters Temporal constraints may be associated with ftUserAttrs aux object class based on:
    user.constraint.name - just a label, i.e. uid
    user.constraint.timeout - 99 - set the integer timeout that contains max time (in minutes) that entity may remain inactive.    
    user.constraint.begin_date - YYYYMMDD - determines date when user may be activated.
    user.constraint.end_date - YYMMDD - indicates latest date user may be activated.
    user.constraint.begin_lock_date - YYYYMMDD - determines beginning of enforced inactive status
    user.constraint.end_lock_date - YYMMDD - determines end of enforced inactive status.
    user.constraint.begin_time - HHMM - determines begin hour user may be activated.
    user.constraint.end_time -  user.role.constraints.end_time -  HHMM - determines end hour user may be activated.
    user.constraint.day_mask - 1234567, 1 = Sunday, 2 = Monday, etc - specifies which day of week user may be activated.
    user.props - multi-occurring name:value pairs
    user.pw_policy - slapd pwpolicy
    
    standard iNetOrgPerson attrs, more info here: https://tools.ietf.org/html/rfc2798:
    user.ou
    user.cn
    user.sn
    user.dn
    user.description
    user.display_name
    user.employee_type
    user.title
    user.phones
    user.mobiles
    user.emails
    user.department_number
    user.l
    user.physical_delivery_office_name
    user.postal_code
    user.room_number                                     
    """    
    utils.validate_user(user)
    return userdao.create(user)

    
def update_user(user):
    """
    This method performs an update on User entity in directory. Prior to making this call the entity must exist in directory.
         
    required parameters:
    user.uid - maps to INetOrgPerson uid     
    optional parameters Temporal constraints may be associated with ftUserAttrs aux object class based on:
    user.constraint.name - just a label, i.e. uid
    user.constraint.timeout - 99 - set the integer timeout that contains max time (in minutes) that entity may remain inactive.    
    user.constraint.begin_date - YYYYMMDD - determines date when user may be activated.
    user.constraint.end_date - YYMMDD - indicates latest date user may be activated.
    user.constraint.begin_lock_date - YYYYMMDD - determines beginning of enforced inactive status
    user.constraint.end_lock_date - YYMMDD - determines end of enforced inactive status.
    user.constraint.begin_time - HHMM - determines begin hour user may be activated.
    user.constraint.end_time -  user.role.constraints.end_time -  HHMM - determines end hour user may be activated.
    user.constraint.day_mask - 1234567, 1 = Sunday, 2 = Monday, etc - specifies which day of week user may be activated.
    user.props - multi-occurring name:value pairs
    user.pw_policy - slapd pwpolicy
    standard iNetOrgPerson attrs, more info here: https://tools.ietf.org/html/rfc2798
    user.ou
    user.cn
    user.sn
    user.dn
    user.description
    user.display_name
    user.employee_type
    user.title
    user.phones
    user.mobiles
    user.emails
    user.department_number
    user.l
    user.physical_delivery_office_name
    user.postal_code
    user.room_number                                     
    """    
    utils.validate_user(user)
    return userdao.update(user)

    
def delete_user(user):
    """
    This command deletes an existing user from the RBAC database. The command is valid if and only if the user to be deleted is a member of the USERS data set. 
    The USERS and UA data sets and the assigned_users function are updated. This method performs a "hard" delete. 
    It completely removes all data associated with this user from the directory. 
    User entity must exist in directory prior to making this call else exception will be thrown.
    
    required parameters:
    user.uid - maps to INetOrgPerson uid     
    """    
    utils.validate_user(user)
    # get the user's role assignments from its entry.
    out_user = userdao.read(user)
    # it's needed to remove the user membership from associated role entries.    
    for role_nm in out_user.roles:
        try:
            roledao.remove_member(Role(name=role_nm), user.uid)
            logger.info('admin.delete_user:' + user.uid + ', removed as member of role:' + role_nm)
        except RbacError as e:
            if e.id != global_ids.URLE_ASSIGN_NOT_EXIST:
                raise RbacError(msg=e.msg, id=e.id)
            else:
                logger.warn('admin.delete_user:' + user.uid + ', is not occupant of role:' + role_nm)
            
    # now remove the user entry:
    return userdao.delete(user)

            
def add_role(role):
    """
    This command creates a new role. The command is valid if and only if the new role is not already a member of the ROLES data set. 
    The ROLES data set is updated. Initially, no user or permission is assigned to the new role.
         
    required parameters:
    role.name - maps to INetOrgPerson uid
    
    optional parameters Temporal constraints may be associated with ftUserAttrs aux object class based on:
    role.props - multi-occurring name:value pairs
    role.description

    role.constraint.timeout - 99 - set the integer timeout that contains max time (in minutes) that entity may remain inactive.        
    role.constraint.beginDate - YYYYMMDD - determines date when role may be activated.
    role.constraint.endDate - YYMMDD - indicates latest date role may be activated.
    role.constraint.beginLockDate - YYYYMMDD - determines beginning of enforced inactive status
    role.constraint.endLockDate - YYMMDD - determines end of enforced inactive status.
    role.constraint.beginTime - HHMM - determines begin hour role may be activated in user's session.
    role.constraint.endTime - HHMM - determines end hour role may be activated in user's session.*
    role.constraint.dayMask - 1234567, 1 = Sunday, 2 = Monday, etc - specifies which day of week role may be activated.
    """    
    utils.validate_role(role)
    return roledao.create(role)

    
def update_role(role):
    """
    Method will update a Role entity in the directory. The role must exist in role container prior to this call. 
    
    required parameters:
    role.name - maps to INetOrgPerson uid
         
    optional parameters Temporal constraints may be associated with ftUserAttrs aux object class based on:
    role.props - multi-occurring name:value pairs
    role.description
    
    role.constraint.timeout - 99 - set the integer timeout that contains max time (in minutes) that entity may remain inactive.    
    role.constraint.beginDate - YYYYMMDD - determines date when role may be activated.
    role.constraint.endDate - YYMMDD - indicates latest date role may be activated.
    role.constraint.beginLockDate - YYYYMMDD - determines beginning of enforced inactive status
    role.constraint.endLockDate - YYMMDD - determines end of enforced inactive status.
    role.constraint.beginTime - HHMM - determines begin hour role may be activated in user's session.
    role.constraint.endTime - HHMM - determines end hour role may be activated in user's session.*
    role.constraint.dayMask - 1234567, 1 = Sunday, 2 = Monday, etc - specifies which day of week role may be activated.
    """    
    utils.validate_role(role)
    return roledao.update(role)

    
def delete_role(role):
    """
    This command deletes an existing role from the RBAC database. The command is valid if and only if the role to be deleted is a member of the ROLES data set. 
    This command will also deassign role from all users.
     
    required parameters:
    role.name - maps to INetOrgPerson uid     
    """    
    utils.validate_role(role)
    
    # if role has members, deassign all.
    members, constraint = roledao.get_members_constraint (role)
    for member in members:
        try:
            userdao.deassign(User(uid=member), constraint)
            logger.info('admin.delete_role:' + role.name + ', remove assign for user:' + member)
        except RbacError as e:
            if e.id != global_ids.URLE_ASSIGN_NOT_EXIST:
                raise RbacError(msg=e.msg, id=e.id)
            else:
                logger.warn('admin.delete_role:' + role.name + ', assign not exist for user:' + member)
            
    # if role is assigned to perms (i.e. granted), remove them as well.
    perms = permdao.search_on_roles([role.name])
    for perm in perms:
        permdao.revoke(perm, role)
        logger.info('admin.delete_role:' + role.name + ', remove grant for perm obj_name:' + perm.obj_name + ', op_name:' + perm.op_name)
    
    # now remove the role entry.                
    return roledao.delete(role)

                        
def add_perm(perm):
    """
    This method will add permission operation to an existing permission object which resides under ou=Permissions,ou=RBAC,dc=yourHostName,dc=com container in directory information tree. 
    The perm operation entity may have Role or User associations. The target Permission must not exist prior to calling. 
    A Fortress Permission instance exists in a hierarchical, one-many relationship between its parent and itself as stored in ldap tree: (PermObj*->Permission).
    
    required parameters:
    perm.obj_name - maps to already existing perm object    
    perm.op_name - accepts an arbitrary name for an operation that maps to runtime process.
        
    optional parameters:
    perm.obj_id - object identifier
    perm.props - multi-occurring property key and values are separated with a ':'. e.g. mykey1:myvalue1
    perm.type - any safe text
    perm.description - any safe text        
    """    
    utils.validate_perm(perm)
    return permdao.create(perm)

    
def update_perm(perm):
    """
    This method will update permission operation pre-existing in target directory under ou=Permissions,ou=RBAC,dc=yourHostName,dc=com container in directory information tree. 
    The perm operation entity may also contain Role or User associations to add or remove using this function. 
    The perm operation must exist before making this call. Only non-null attributes will be updated.

    required parameters:
    perm.obj_name - maps to already existing perm object    
    perm.op_name - maps to already existing op name
        
    optional parameters:
    perm.obj_id - object identifier
    perm.props - multi-occurring property key and values are separated with a ':'. e.g. mykey1:myvalue1
    perm.type - any safe text
    perm.description - any safe text                
    """    
    utils.validate_perm(perm)
    return permdao.update(perm)

    
def delete_perm(perm):
    """
    This method will remove permission operation entity from permission object. A Fortress permission is (object->operation). The perm operation must exist before making this call.
    
    required parameters:
    perm.obj_name - maps to already existing perm object    
    perm.op_name - maps to already existing op name
                
    optional parameters:
    perm.obj_id    
    """    
    utils.validate_perm(perm)
    return permdao.delete(perm)

                                                
def add_object(perm_obj):
    """
    This method will add permission object to perms container in directory. The perm object must not exist before making this call. 
    A PermObj instance exists in a hierarchical, one-many relationship between itself and children as stored in ldap tree: (PermObj*->Permission).
    
    required parameters:
    perm.obj_name - maps to arbitrary name of system resource.
            
    optional parameters
    perm_obj.description - any safe text
    perm_obj.type - contains any safe text
    perm_obj.props * - multi-occurring property key and values are separated with a ':'. e.g. mykey1:myvalue1
    """    
    utils.validate_perm_obj(perm_obj)
    return permdao.create_obj(perm_obj)

    
def update_object(perm_obj):
    """
    This method will update permission object in perms container in directory. 
    The perm object must exist before making this call. 
    A PermObj instance exists in a hierarchical, one-many relationship between itself and children as stored in ldap tree: (PermObj*->Permission).
        
    required parameters:
    perm.obj_name - maps to existing perm object.
            
    optional parameters
    perm_obj.description - any safe text
    perm_obj.type - contains any safe text
    perm_obj.props * - multi-occurring property key and values are separated with a ':'. e.g. mykey1:myvalue1
    """    
    utils.validate_perm_obj(perm_obj)
    return permdao.update_obj(perm_obj)

    
def delete_object(perm_obj):
    """
    This method will remove permission object to perms container in directory. This method will also remove in associated permissions that are attached to this object.
    
    required parameters:
    perm.obj_name - maps to existing perm object.        
    """    
    utils.validate_perm_obj(perm_obj)
    try:
        permdao.delete_obj(perm_obj)
    except RbacError as e:
        # if entry has children.
        if e.id == global_ids.PERM_OBJECT_DELETE_FAILED_NONLEAF:
            logger.warn('admin.delete_object non-leaf, obj_name:' + perm_obj.obj_name)
            # remove all of them.
            pList = permdao.search(Perm(obj_name=perm_obj.obj_name, op_name='*'))
            for perm in pList:
                permdao.delete(perm)
                logger.warn('admin.delete_object child obj_name:' + perm.obj_name + ', op_name:' + perm.op_name)
                
            # now try to remove this node once again
            permdao.delete_obj(perm_obj)
            logger.warn('admin.delete_object success after retry, obj_name:' + perm.obj_name)
        else:
            # can't handle this error so rethrow.
            raise RbacError(msg=e.msg, id=e.id)
    return


def assign(user, role):
    """
    This command assigns a user to a role.

    The command is valid if and only if:
        The user is a member of the USERS data set
        The role is a member of the ROLES data set
        The user is not already assigned to the role

    required parameters:
    user.uid - existing user.        
    role.name - existing role.    
    """    
    utils.validate_user(user)
    utils.validate_role(role)
    entity = roledao.read(role)
    # LDAP doesn't do well with sub-string indexes which is why the role assignments are stored within two separate multi-occurring attributes on the user entry -- roles' and 'role_constraints'.
    # The first, is a set of role names (only), and will be indexed for fast search.
    # the second, is a set of delimited strings containing the role name (once again) plus its associated temporal values. 
    userdao.assign(user, entity.constraint)
    
    # Fortress user-role assignments also keep member association on the role itself. 
    # The rationale for these assignments also stored on role is two-fold:
    # 1. works with traditional LDAP group-based authZ mechanisms
    # 2. makes role-users search query more efficient, as its stored on single entry.     
    roledao.add_member(entity, user.uid)

                                                                                                
def deassign(user, role):
    """
    This command deletes the assignment of the User from the Role entities. 
    The command is valid if and only if the user is a member of the USERS data set, the role is a member of the ROLES data set, 
    and the user is assigned to the role. Any sessions that currently have this role activated will not be effected. 
    Successful completion includes:
    User entity in USER data set has role assignment removed.
    Role entity in ROLE data set has userId removed as role occupant.
    
    required parameters:
    user.uid - existing user.        
    role.name - existing role.        
    """    
    utils.validate_user(user)
    utils.validate_role(role)
    out_user = userdao.read(user)
    for constraint in out_user.role_constraints:
        if constraint.name == role.name:
            found = True
            userdao.deassign(user, constraint)
            try:
                roledao.remove_member(role, user.uid)
            except RbacError as e:
                if e.id != global_ids.URLE_ASSIGN_NOT_EXIST:
                    raise RbacError(msg=e.msg, id=e.id)
                else:
                    logger.warn('admin.deassign remove member failed because not occupant. user:' + user.uid + ', role:' + role.name)
    if not found:
        raise RbacError(msg='Role deassign failed constraint not found', id=global_ids.URLE_DEASSIGN_FAILED)
        
            
def grant(perm, role):
    """
    This method will add permission operation to an existing permission object which resides under ou=Permissions,ou=RBAC,dc=yourHostName,dc=com container in directory information tree. 
    The perm operation entity may have Role or User associations. The target Permission must not exist prior to calling. 
    A Fortress Permission instance exists in a hierarchical, one-many relationship between its parent and itself as stored in ldap tree: (PermObj*->Permission).
        
    required parameters:
    perm.obj_name - existing perm obj.
    perm.obj_name - existing perm op.                
    role.name - existing role.

    optional parameters:
    perm.obj_id - object identifier
    """    
    utils.validate_role(role)
    utils.validate_perm(perm)
    return permdao.grant(perm, role)

                                                                                                
def revoke(perm, role):
    """
    This command revokes the permission to perform an operation on an object from the set of permissions assigned to a role. 
    The command is implemented by setting the access control list of the object involved. 
    The command is valid if and only if the pair (operation, object) represents a permission, the role is a member of the ROLES data set, and the permission is assigned to that role.
    
    required parameters:
    perm.obj_name - existing perm obj.
    perm.obj_name - existing perm op.                
    role.name - existing role.
        
    optional parameters:
    perm.obj_id - object identifier
    """    
    utils.validate_role(role)
    utils.validate_perm(perm)
    return permdao.revoke(perm, role)
                                                                                                                                                                         


'''
@copyright: 2022 - Symas Corporation
'''

from rbac.model import Session
from rbac.util import Date, Day, LockDate, Time, TimeOut, CurrentDateTime
from rbac.ldap import userdao, permdao
from rbac.util.fortress_error import RbacError
from rbac.util import logger, global_ids, SUCCESS


def create_session (user, is_trusted):
    """
    Perform user authentication User.password and role activations.
    This method must be called once per user prior to calling other methods within this module. 
    The successful result is Session that contains target user's RBAC User.roles.
    
    This API will...

    * authenticate user password if trusted == false.
    * evaluate temporal Constraint(s) on User and UserRoles.
    * process selective role activations into User RBAC Session User.roles.
    * return a Session containing Session.user, Session.user.roles
    
    required parameters:
    user.uid - maps to INetOrgPerson uid
    is_trusted - boolean, if 'True', authentication is skipped (password not checked)
    
    return:
    Session     
    """    
    __validate_user(user)
    session = Session()
    if is_trusted is False:
        # failure throws exception:
        userdao.authenticate(user)
        session.is_authenticated = True
    entity = userdao.read(user)
    session.user = entity            
    __validate_user_constraint(session, 'create_session')
    __validate_role_constraints(session)
    __refresh(session)
    return session


def check_access (session, perm):
    """
    Perform user RBAC authorization. 
    This function returns a Boolean value meaning whether the subject of a given session is allowed or not to perform a given operation on a given object. 
    The function is valid if and only if the session is a valid Fortress session,
    the object is a member of the OBJS data set, and the operation is a member of the OPS data set.
    The session's subject has the permission to perform the operation on that object if and only if that permission is assigned to (at least) one of the session's active roles.
    This implementation will verify the roles or userId correspond to the subject's active roles are registered in the object's access control list.
    
    required parameters:
    session - as returned from create_session api
    perm.obj_name - maps to already existing perm object    
    perm.op_name - maps to already existing op name    
        
    optional parameters:
    perm.obj_id    
    
    return:
    boolean True if allowed; False otherwise     
    """        
    __validate(session)
    __validate_perm(perm)    
    __validate_user_constraint(session, 'check_access')
    __validate_role_constraints(session)    
    result = False
    entity = permdao.read(perm)
    for role in session.user.roles:
        if __is_role_found(role, entity.roles):        
            result = True
            break
    __refresh(session)
    return result


def is_user_in_role (session, role):
    """
    This function returns a BOOLEAN value whether subject has the specified role contained within their session.
         
    required parameters:
    session - as returned from create_session api    
    role.name - maps to existing role     
    
    return:
    boolean True if allowed; False otherwise         
    """        
    __validate(session)
    __validate_user_constraint(session, 'is_user_in_role')    
    result = False
    __validate_role_constraints(session)
    if __is_role_found(role, session.user.roles):
        result = True
    __refresh(session)
    return result


def add_active_role (session, role):
    """
    This function adds a role as an active role of a session whose owner is a given user. 
    
    required parameters:
    session - as returned from create_session api    
    role.name - maps to existing role     
    
    return:
    None     
    """    
    __validate(session) 
    __validate_user_constraint(session, 'add_active_role')   
    if any ( s.lower() == role.lower() for s in session.user.roles ):
        raise RbacError (msg='add_active_role uid=' + session.user.uid + ', previously activated role=' + role, id=global_ids.URLE_ALREADY_ACTIVE)
    user = userdao.read(session.user)        
    constraint = __find_role_constraint(role, user.role_constraints)
    if constraint is not None:
       __activate_role(session.user, constraint)
    else:
        raise RbacError (msg='add_active_role uid=' + session.user.uid + ', has not been assigned role=' + role, id=global_ids.URLE_ASSIGN_NOT_EXIST)
    __validate_role_constraints(session)
    __refresh(session)


def drop_active_role (session, role):
    """
    This function deletes a role from the active role set of a session owned by a given user. 
    The function is valid if and only if the user is a member of the USERS data set, the session object contains a valid Fortress session, 
    the session is owned by the user, and the role is an active role of that session.
    
    required parameters:
    session - as returned from create_session api    
    role.name - maps to existing role     
    
    return:
    None     
    """    
    __validate(session)
    __validate_user_constraint(session, 'drop_active_role')
    constraint = __find_role_constraint(role, session.user.role_constraints)
    if constraint is not None:
        __deactivate_role(session.user, constraint)
    else:
        raise RbacError (msg='drop_active_role uid=' + session.user.uid + ', has not activated role=' + role, id=global_ids.URLE_NOT_ACTIVE)
    __validate_role_constraints(session)
    __refresh(session)


def session_roles (session):
    """
    This function returns the active roles associated with a session. The function is valid if and only if the session is a valid Fortress session.

    required parameters:        
    session - as returned from create_session api    
    
    return:
    Constraint list     
    """    
    __validate(session)
    __validate_user_constraint(session, 'session_roles')
    __validate_roles(session.user)    
    __validate_role_constraints(session)
    __refresh(session)
    return session.user.role_constraints            


def session_perms (session):
    """
    This function returns the permissions of the session, i.e., the permissions assigned to its authorized roles. 
    The function is valid if and only if the session is a valid Fortress session.

    required parameters:    
    session - as returned from create_session api    
    
    return:
    Perm list     
    """    
    __validate(session)
    __validate_user_constraint(session, 'session_perms')
    __validate_roles(session.user)    
    __validate_role_constraints(session)
    __refresh(session)            
    return permdao.search_on_roles(session.user.roles)


def __activate_role(user, role_constraint):
    user.roles.append(role_constraint.name)
    user.role_constraints.append(role_constraint)


def __deactivate_role(user, role_constraint):
    user.roles.remove(role_constraint.name)
    user.role_constraints.remove(role_constraint)


def __refresh(session):
    session.last_access = CurrentDateTime()            


def __validate(session):
    if session is None:
        raise RbacError (msg='Session is None', id=global_ids.USER_SESS_NULL)
    elif session.user is None:
        raise RbacError (msg='Session has no user', id=global_ids.USER_SESS_NULL)


def __validate_roles(user):
    if user.roles is None:
        raise RbacError (msg='User roles is None', id=global_ids.ROLE_LST_NULL)
    elif len(user.roles) < 1:
        raise RbacError (msg='User roles is Empty', id=global_ids.ROLE_LST_NULL)


def __validate_user(user):
    if user is None:
        raise RbacError (msg='User is None', id=global_ids.USER_NULL)
    elif user.uid is None:
        raise RbacError (msg='User uid is None', id=global_ids.USER_ID_NULL)


def __validate_perm(perm):
    if perm is None:
        raise RbacError (msg='Perm is None', id=global_ids.PERM_NULL)
    elif perm.obj_name is None:
        raise RbacError (msg='Perm object name is None', id=global_ids.PERM_OBJECT_NM_NULL)
    elif perm.op_name is None:
        raise RbacError (msg='Perm op name is None', id=global_ids.PERM_OPERATION_NM_NULL)


def __validate_role_constraints(session):
    for role_constraint in session.user.role_constraints:
        result = __validate_role_constraint(session, role_constraint)
        if result is not SUCCESS:
                logger.debug('validate_role_constraints deactivate user-role:' + session.user.uid + '.' + role_constraint.name + ', result=' + str(result))
                __deactivate_role(session.user, role_constraint)                


def __is_constraint(constraint):
    is_valid = True
    if constraint.raw is not None and not constraint.raw:
        is_valid = False
    return is_valid


def __validate_user_constraint(session, op):
    result = SUCCESS
    if __is_constraint(session.user.constraint):
        for validator in validators:
            result = validator.validate(session.user.constraint, CurrentDateTime(), session)
            if result is not SUCCESS:
                logger.debug(validator.__class__.__name__ + ' validation failed:' + session.user.constraint.name + ', result=' + str(result))
                raise RbacError (msg=op + ' constraint validation failed uid:' + session.user.uid, id=result)
    return result


def __validate_role_constraint(session, constraint):
    result = SUCCESS
    if __is_constraint(constraint):
        for validator in validators:
            result = validator.validate(constraint, CurrentDateTime(), session)
            if result is not SUCCESS:
                logger.debug(validator.__class__.__name__ + ' validation failed:' + constraint.name + ', uid=' + session.user.uid + ', result=' + str(result))
                break
    return result


def __is_role_found(role, roles):
    result = False
    if any ( s.lower() == role.lower() for s in roles ):        
        result = True
    return result


def __find_role_constraint(role, role_constraints):
    result = None    
    for role_constraint in role_constraints:
        if role.lower() == role_constraint.name.lower():
            result = role_constraint
    return result


# Initialize the constraint validators:
validators = []
validators.append(Date())
validators.append(Day())
validators.append(LockDate())
validators.append(Time())
validators.append(TimeOut())


'''
@copyright: 2022 - Symas Corporation
'''

from .validator import Validator
from .global_ids import ACTV_FAILED_TIME, SUCCESS

class Time(Validator):
        
    def validate(self, constraint, now, session=None):
        if not constraint.begin_time or constraint.begin_time is None or not constraint.end_time or constraint.end_time is None:
            return SUCCESS        
        elif constraint.begin_time == '0000' and constraint.end_time == '0000':
            return SUCCESS        
        elif constraint.begin_time <= now.time <= constraint.end_time:
            return SUCCESS
        else:
            return ACTV_FAILED_TIME        



'''
@copyright: 2022 - Symas Corporation
'''

import unittest
import inspect

from rbac import (
    # model
    User,
    Role,
    Perm,
    PermObj,
    # apis:
    review,
    admin,
    access,
    #exception handling:
    RbacError,
    global_ids
)


class BasicTestSuite(unittest.TestCase):
    """These tests the py-fortress review functions."""
                

class TestSamples(unittest.TestCase):
    """
    Basic tests for beginners
    """    

    def test22_read_user(self):
        """
        Read a user that was created earlier. Expects a unique uid that points to an existing user.
        """
        print_test_name()
        try:
            out_user = review.read_user(User(uid='foo1'))
            print_user(out_user)
        except RbacError as e:            
            print_exception(e)
            self.fail()


    def test23_search_users(self):
        """
        Search for users that match the characters passed into with wildcard appended.  Will return zero or more records, one for each user in result set.
        """
        print_test_name()
        try:
            users = review.find_users(User(uid='foo*'))
            for user in users:
                print_user(user)
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test24_assigned_users(self):
        """
        Search for users that are assigned a particular role.  Will return zero or more records, one for each user in result set.
        """
        print_test_name()
        try:
            uids = review.assigned_users(Role(name='Customer'))
            for uid in uids:
                print_test_msg('uid=' + uid)
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test25_perm_users(self):
        """
        Search for users that have been authorized a particular permission.  Will return zero or more records, of type user, one for each user authorized that particular perm.
        """
        print_test_name()
        try:
            users = review.perm_users(Perm(obj_name='ShoppingCart', op_name='add'))
            for user in users:
                print_user(user)
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test10_add_user(self):
        """
        To add a user in py-fortress, the uid must be supplied.  All other attributes are optional but to be able to authenticate a password must be present of course.
        """
        print_test_name()
        
        try:
            admin.add_user(User(uid='foo1', password='secret'))
            print_test_msg('success')                        
        except RbacError as e:
            print_exception(e)
            self.fail()

    def test01_delete_user(self):
        """
        The delete expects the uid passed in to match one record only and will throw an exception if not found, or otherwise if error occurs.
        """
        print_test_name()
        try:
            admin.delete_user(User(uid='foo1'))
            print_test_msg('success')           
        except RbacError as e:
            if e.id == global_ids.USER_NOT_FOUND:
                print_test_msg('not found')
                pass
            else:
                print_exception(e)
                self.fail()


    def test12_read_role(self):
        """
        The read role expects the role name to point to an existing entry and will throw an exception if not found or other error occurs.
        """
        print_test_name()        
        try:
            out_role = review.read_role(Role(name='Customer'))
            print_role(out_role)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test14_search_roles(self):
        """
        Search for roles that match the characters passed into with wildcard appended.  Will return zero or more records, one for each user in result set.
        """
        print_test_name()        
        try:
            roles = review.find_roles(Role(name='Customer*'))
            for role in roles:
                print_role(role)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test15_assigned_roles(self):
        """
        Return the list of roles that have been assigned a particular user.  Will return zero or more records, of type constraint, one for each role assigned to user.
        """
        print_test_name()        
        try:
            constraints = review.assigned_roles(User(uid='foo1'))
            for constraint in constraints:
                print_test_msg('role name=' + constraint.name)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test16_perm_roles(self):
        """
        Return the list of roles that have granted a particular perm.  Will return zero or more records, containing the role names, one for each role assigned to permission.
        """
        print_test_name()        
        try:
            names = review.perm_roles(Perm(obj_name='ShoppingCart', op_name='add'))
            for name in names:
                print_test_msg('role name=' + name)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test08_add_role(self):
        """
        A fortress role only need to be supplied a name but it must be unique, and not present.
        """
        print_test_name()        
        try:
            admin.add_role(Role(name='Customer'))
            print_test_msg('success')                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test05_delete_role(self):
        """
        The delete will throw an exception if the passed role is not found or otherwise fails to process.
        """
        print_test_name()        
        try:
            admin.delete_role(Role(name='Customer'))
            print('test5_delete_role success')
        except RbacError as e:
            if e.id == global_ids.ROLE_NOT_FOUND:
                print('test5_delete_role not found')
                pass
            else:            
                print_exception(e)
                self.fail()


    def test13_read_obj(self):
        """
        The ob_name is the only required attribute on a fortress object. Will throw an exception if not found.
        """
        print_test_name()        
        try:
            out_obj = review.read_object(PermObj(obj_name='ShoppingCart'))
            print_obj(out_obj)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test17_search_objs(self):
        """
        Search for ojects that match the characters passed into with wildcard appended.  Will return zero or more records, one for each user in result set.
        """
        print_test_name()        
        try:
            objs = review.find_objects(PermObj(obj_name='ShoppingCart*'))
            for obj in objs:
                print_obj(obj)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test06_add_obj(self):
        """
        The ob_name is the only required attribute on a fortress object. Will throw an exception if not unique.
        """
        print_test_name()
        
        try:
            admin.add_object(PermObj(obj_name='ShoppingCart'))
            print_test_msg('success')                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test04_delete_obj(self):
        """
        Delete a basic perm object
        """
        print_test_name()
        
        try:
            admin.delete_object(PermObj(obj_name='ShoppingCart'))
            print_test_msg('success')                        
        except RbacError as e:
            if e.id == global_ids.PERM_OBJ_NOT_FOUND:
                print_test_msg('obj not found')
                pass
            else:            
                print_exception(e)
                self.fail()


    def test18_read_perm(self):
        """
        Permissions require obj_name and op_name, obj_id is optional.  This will throw an exception if not found.
        """
        print_test_name()        
        try:
            out_perm = review.read_perm(Perm(obj_name='ShoppingCart', op_name='add'))
            print_perm(out_perm)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test19_search_perms(self):
        """
        Search for perms that match the characters passed into with wildcard appended.  Will return zero or more records, one for each user in result set.
        """
        print_test_name()        
        try:
            perms = review.find_perms(Perm(obj_name='ShoppingCart*', op_name='*'))
            for perm in perms:
                print_perm(perm)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test20_role_perms(self):
        """
        Search for perms that have been granted to a particular role.  Will return zero or more records, of type permission, one for each grant.
        """
        print_test_name()        
        try:
            perms = review.role_perms(Role(name='Customer'))
            for perm in perms:
                print_perm(perm)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test21_user_perms(self):
        """
        Search for perms that have been authorized to a particular user based on their role assignments.  Will return zero or more records, of type permission, one for each perm authorized for user.
        """
        print_test_name()        
        try:
            perms = review.user_perms(User(uid='foo1'))
            for perm in perms:
                print_perm(perm)                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test07_add_perm(self):
        """
        Permissions require obj_name and op_name, obj_id is optional.  This will throw an exception if not unique.
        """
        print_test_name()
        
        try:
            admin.add_perm(Perm(obj_name='ShoppingCart', op_name='add'))
            print_test_msg('success')                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test03_delete_perm(self):
        """
        Permissions require obj_name and op_name, obj_id is optional.  This will throw an exception if not found.
        """
        print('test3_delete_perm')
        try:
            admin.delete_perm(Perm(obj_name='ShoppingCart', op_name='add'))
            print_test_msg('success')                        
        except RbacError as e:
            if e.id == global_ids.PERM_OP_NOT_FOUND:
                print_test_msg('not found')
                pass
            else:            
                print_exception(e)
                self.fail()


    def test09_grant_perm(self):
        """
        Grant a permission to a role passing required attrs for perm (obj_name op_name) and role (name). Throws exception if perm not found, or already have been granted.
        """
        print_test_name()
        
        try:
            admin.grant(Perm(obj_name='ShoppingCart', op_name='add'), Role(name="Customer"))
            print_test_msg('success')                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test02_revoke_perm(self):
        """
        Revokes a permission from a role passing required attrs for perm (obj_name op_name) and role (name). Throws exception if perm not found, or if not granted.
        """
        print_test_name()
        
        try:
            admin.revoke(Perm(obj_name='ShoppingCart', op_name='add'), Role(name="Customer"))
            print_test_msg('success')                        
        except RbacError as e:
            if e.id == global_ids.PERM_NOT_EXIST:
                print_test_msg('not granted')
            else:
                print_exception(e)
                self.fail()


    def test11_assign_user(self):
        """
        Assign a user to a role passing req's attrs for user (uid) and role (name). Throws exception if either are not found, or already assigned.
        """
        print_test_name()
        
        try:
            admin.assign(User(uid='foo1'), Role(name='Customer'))
            print_test_msg('success')                        
        except RbacError as e:
            print_exception(e)
            self.fail()


    def test26_create_session(self):
        """
        Called when beginning an RBAC session. Requires uid and password if trusted is False, in which case it will authenticate in addition to activating roles into session.
        Throws exception if user is not found, or if authentication fails. 
        """
        print_test_name()
        
        try:
            session = access.create_session(User(uid='foo1', password='secret'), False)
            if not session:
                print_test_msg('fail')
                self.fail()
            else:
                print_test_msg('sucess')
                pass                        
        except RbacError as e:
            print_exception(e)
            self.fail()
                        

    def test27_check_access(self):
        """
        Called to perform an RBAC permmission check.  Requires a valid session (returned from create_session) and a permission with required attrs obj_name, op_name, and optional obj_id.
        Returns true if user is allowed to perform the specified operation on object.
        """
        print_test_name()
        
        try:
            session = access.create_session(User(uid='foo1', password='secret'), False)
            result = access.check_access(session, Perm(obj_name='ShoppingCart', op_name='add'))
            if not result:
                print_test_msg('fail')
                self.fail()
            else:
                print_test_msg('success')
                pass                        
        except RbacError as e:
            print_exception(e)
            self.fail()
                        

    def test28_session_perms(self):
        """
        Return all permission authorized for user per their set of activated roles in the session.  Note this means that set could be a subset of assigned permissions, if any assigned roles are not activated.
        """
        print_test_name()
        
        try:
            session = access.create_session(User(uid='foo1', password='secret'), False)
            perms = access.session_perms(session)
            if not perms:
                print_test_msg('failed')
                self.fail()
            
            for perm in perms:
                print_perm(perm)
            pass                        
        except RbacError as e:
            print_exception(e)
            self.fail()
            
            
def print_test_name(): 
    print(inspect.currentframe().f_back.f_code.co_name)
        

def print_test_msg(msg=None):
    if msg is None:
        msg = '' 
    print(inspect.currentframe().f_back.f_code.co_name + ' ' + msg)
        

def print_exception(e):
    print(inspect.currentframe().f_back.f_code.co_name + ' exception id=' + str(e.id) + ', msg=' + e.msg)
        

def print_user (entity):
        caller = inspect.currentframe().f_back.f_code.co_name
        print(caller + ' uid:' + entity.uid)

        
def print_role (entity):
        caller = inspect.currentframe().f_back.f_code.co_name
        print(caller + ' role name:' + entity.name)


def print_obj (entity):
        caller = inspect.currentframe().f_back.f_code.co_name
        print(caller + ' object name:' + entity.obj_name)


def print_perm (entity):
        caller = inspect.currentframe().f_back.f_code.co_name
        print(caller + ' object name:' + entity.obj_name + ' op name:' + entity.op_name)    
        
        
def suite():
    suite = unittest.TestSuite()

    # Teardown:        
    suite.addTest(TestSamples('test01_delete_user'))
    suite.addTest(TestSamples('test02_revoke_perm'))
    suite.addTest(TestSamples('test03_delete_perm'))
    suite.addTest(TestSamples('test04_delete_obj'))
    suite.addTest(TestSamples('test05_delete_role'))
        
    # Buildup:
    suite.addTest(TestSamples('test06_add_obj'))
    suite.addTest(TestSamples('test07_add_perm'))
    suite.addTest(TestSamples('test08_add_role'))
    suite.addTest(TestSamples('test09_grant_perm'))
    suite.addTest(TestSamples('test10_add_user'))
    suite.addTest(TestSamples('test11_assign_user'))

    # Interrogate entities and their relationships:
    suite.addTest(TestSamples('test12_read_role'))
    suite.addTest(TestSamples('test13_read_obj'))
    suite.addTest(TestSamples('test14_search_roles'))
    suite.addTest(TestSamples('test15_assigned_roles'))
    suite.addTest(TestSamples('test16_perm_roles'))
    suite.addTest(TestSamples('test17_search_objs'))
    suite.addTest(TestSamples('test18_read_perm'))
    suite.addTest(TestSamples('test19_search_perms'))
    suite.addTest(TestSamples('test20_role_perms'))
    suite.addTest(TestSamples('test21_user_perms'))
    suite.addTest(TestSamples('test22_read_user'))
    suite.addTest(TestSamples('test23_search_users'))
    suite.addTest(TestSamples('test24_assigned_users'))
    suite.addTest(TestSamples('test25_perm_users'))
    
    # Test the policies:    
    suite.addTest(TestSamples('test26_create_session'))
    suite.addTest(TestSamples('test27_check_access'))
    suite.addTest(TestSamples('test28_session_perms'))
    return suite  
 
if __name__ == '__main__':
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())


'''
@copyright: 2022 - Symas Corporation
'''

import unittest
from rbac.ldap import ditdao
from rbac.util import global_ids
from rbac.cli.utils import print_ln


class BasicTestSuite(unittest.TestCase):
    """These tests the py-fortress DIT add and del functions."""
                

class TestDitDao(unittest.TestCase):
    """
    Test the DIT dao funcs
    """    
                     
                      
    def test01_del_containers(self):
        """
        Test the delete ou method
        """
        print_ln('test_del_containers')

        try:
            ditdao.delete_ou(global_ids.USER_OU)
            ditdao.delete_ou(global_ids.ROLE_OU)
            ditdao.delete_ou(global_ids.PERM_OU)
            print('test_del_containers success')
        except Exception as e:
            error = 'test_del_containers errpr=' + e.msg
            print(error)
            #self.fail(error)


    def test02_del_suffix(self):
        """
        Test the del suffix method
        """
        print_ln('test_del_suffix')

        try:
            ditdao.delete_suffix()
            print('test_del_suffix success')
        except Exception as e:
            error = 'test_del_suffix errpr=' + e.msg
            print(error)
            #self.fail(error)


    def test03_bootstrap(self):
        """
        Create the DIT based on the config
        """
        print_ln('test_bootstrap')
        try:
            ditdao.bootstrap()            
            print('test_bootstrap success')                                
        except Exception as e:
            error = 'test_bootstrap errpr=' + e.msg
            print(error)
            #self.fail(error)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestDitDao('test01_del_containers'))
    suite.addTest(TestDitDao('test02_del_suffix'))
    suite.addTest(TestDitDao('test03_bootstrap'))
    return suite

 
def main():
    # My code here
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())


if __name__ == "__main__":
    main()
    


'''
@copyright: 2022 - Symas Corporation
'''

import uuid    
import ldap
from ldap import MOD_REPLACE, MOD_ADD, MOD_DELETE
from ldap.cidict import cidict as CIDict
from rbac.model import Perm, PermObj
from rbac.ldap import ldaphelper, NotFound, NotUnique
from rbac.ldap.ldaphelper import add_to_modlist
from rbac.util import global_ids
from rbac.util.fortress_error import RbacError


def read (entity):
    permList = search(entity)
    if permList is None or len(permList) == 0:
        raise NotFound(msg="Perm Read not found, obj name=" + entity.obj_name + ', op name=' + entity.op_name, id=global_ids.PERM_NOT_EXIST)    
    elif len(permList) > 1:
        raise NotUnique(msg="Perm Read not unique, obj name=" + entity.obj_name + ', op name=' + entity.op_name, id=global_ids.PERM_READ_OP_FAILED)
    else:
        return permList[0]


def search (entity):
    conn = None            
    permList = []
    search_filter = '(&(objectClass=' + PERM_OC_NAME + ')'
    if entity.obj_name:
        search_filter += '(' + OBJ_NM + '=' + entity.obj_name + ')'
    if entity.op_name:
        search_filter += '(' + OP_NM + '=' + entity.op_name + ')'
    if entity.obj_id:
        search_filter += '(' + OBJ_ID + '=' + entity.obj_id + ')'
    search_filter += ')'           
    try:
        conn = ldaphelper.open()
        entries = conn.search_s(__CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=SEARCH_ATTRS)
        for dn, attrs in entries:
            permList.append(__unload(dn, attrs))
    except Exception as e:
        raise RbacError(msg='Perm search error=' + str(e), id=global_ids.PERM_SEARCH_FAILED)
    finally:
        if conn:        
            ldaphelper.close(conn)
    return permList


def read_obj (entity):
    objList = search_objs(entity)
    if objList is None or len(objList) == 0:
        raise NotFound(msg="PermObj Read not found, obj name=" + entity.obj_name, id=global_ids.PERM_OBJ_NOT_FOUND)    
    elif len(objList) > 1:
        raise NotUnique(msg="PermObj Read not unique, obj name=" + entity.obj_name, id=global_ids.PERM_READ_OBJ_FAILED)
    else:
        return objList[0]


def search_objs (entity):
    conn = None            
    permList = []
    search_filter = '(&(objectClass=' + PERM_OBJ_OC_NAME + ')'
    search_filter += '(' + OBJ_NM + '=' + entity.obj_name + '))'           
    try:
        conn = ldaphelper.open()
        entries = conn.search_s(__CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=SEARCH_ATTRS)
        for dn, attrs in entries:
            permList.append(__unload_obj(dn, attrs))
    except Exception as e:
        raise RbacError(msg='PermObj search error=' + str(e), id=global_ids.PERM_SEARCH_FAILED)
    finally:
        if conn:        
            ldaphelper.close(conn)
    return permList


# assumes that roles contains at least one role name
def search_on_roles (roles):    
    conn = None            
    permList = []    
    search_filter = '(&(objectClass=' + PERM_OC_NAME + ')'
    if len (roles) > 1:
        search_filter += '(|'
        end_filter = '))'
    else:
        end_filter = ')'
    for role in roles:
        search_filter += '(' + ROLES + '=' + role + ')'
    search_filter += end_filter                    
    try:
        conn = ldaphelper.open()
        entries = conn.search_s(__CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=SEARCH_ATTRS)
        for dn, attrs in entries:
            permList.append(__unload(dn, attrs))
    except Exception as e:
        raise RbacError(msg='Perm Search Roles error=' + str(e), id=global_ids.PERM_ROLE_SEARCH_FAILED)
    finally:
        if conn:        
            ldaphelper.close(conn)
    return permList


def __unload(dn, attrs):
    entity = Perm()
    entity.dn = dn

    attrs = CIDict(attrs)

    entity.internal_id = ldaphelper.get_attr_val(attrs.get(global_ids.INTERNAL_ID, []))
    entity.obj_id = ldaphelper.get_attr_val(attrs.get(OBJ_ID, []))
    entity.obj_name = ldaphelper.get_attr_val(attrs.get(OBJ_NM, []))
    entity.op_name = ldaphelper.get_attr_val(attrs.get(OP_NM, []))
    entity.abstract_name = ldaphelper.get_attr_val(attrs.get(PERM_NAME, []))
    entity.type = ldaphelper.get_attr_val(attrs.get(TYPE, []))
    entity.description = ldaphelper.get_one_attr_val(attrs.get(global_ids.DESC, []))
    # Get the multi-occurring attrs:
    entity.users = ldaphelper.get_list(attrs.get(USERS, []))
    entity.roles = ldaphelper.get_list(attrs.get(ROLES, []))
    entity.props = ldaphelper.get_list(attrs.get(global_ids.PROPS, []))
    return entity


def __unload_obj(dn, attrs):
    entity = PermObj()

    attrs = CIDict(attrs)

    entity.dn = dn
    entity.internal_id = ldaphelper.get_attr_val(attrs.get(global_ids.INTERNAL_ID, []))
    entity.obj_name = ldaphelper.get_attr_val(attrs.get(OBJ_NM, []))
    entity.type = ldaphelper.get_attr_val(attrs.get(TYPE, []))
    entity.description = ldaphelper.get_one_attr_val(attrs.get(global_ids.DESC, []))
    entity.ou = ldaphelper.get_one_attr_val(attrs.get(global_ids.OU, []))
    entity.props = ldaphelper.get_list(attrs.get(global_ids.PROPS, []))
    return entity


def create ( entity ):
    try:
        attrs = {}
        attrs.update( {'objectClass': PERM_OCS} )
        attrs.update( {OBJ_NM : entity.obj_name} )
        attrs.update( {OP_NM : entity.op_name} )
        entity.abstract_name = entity.obj_name + '.' + entity.op_name        
        attrs.update( {global_ids.CN : entity.abstract_name} )
                
        # generate random id:
        entity.internal_id = str(uuid.uuid4())
        attrs.update( {global_ids.INTERNAL_ID : entity.internal_id} )
        if entity.obj_id:        
            attrs.update( {OBJ_ID : entity.obj_id} )
        if entity.description:        
            attrs.update( {global_ids.DESC : entity.description} )
        if entity.abstract_name:        
            attrs.update( {PERM_NAME : entity.abstract_name} )
        if entity.type:        
            attrs.update( {TYPE : entity.type} )
        # list of strings:
        if entity.props is not None and len(entity.props) > 0 :        
            attrs.update( {global_ids.PROPS : entity.props} )            
        if entity.users is not None and len(entity.users) > 0 :        
            attrs.update( {USERS : entity.users} )        
        if entity.roles is not None and len(entity.roles) > 0 :        
            attrs.update( {ROLES : entity.roles} )
        conn = ldaphelper.open()        
        conn.add_s(__get_dn(entity), add_to_modlist(attrs))
    except ldap.ALREADY_EXISTS:
        raise NotUnique(msg='Perm create failed, already exists:' + entity.obj_name, id=global_ids.PERM_ADD_FAILED)
    except ldap.LDAPError as e:
        raise RbacError(msg='Perm create failed result=' + str(e), id=global_ids.PERM_ADD_FAILED)
    except Exception as e:
        raise RbacError(msg='Perm create error=' + str(e), id=global_ids.PERM_ADD_FAILED)
    return entity


def update ( entity ):
    try:
        attrs = {}
        if entity.description:        
            attrs.update( {global_ids.DESC : [(MOD_REPLACE, [entity.description])]} )
        if entity.type:        
            attrs.update( {TYPE : [(MOD_REPLACE, [entity.type])]} )
        # list of strings:
        if entity.props is not None and len(entity.props) > 0 :        
            attrs.update( {global_ids.PROPS : [(MOD_REPLACE, entity.props)]} )
        if entity.users is not None and len(entity.users) > 0 :        
            attrs.update( {USERS : [(MOD_REPLACE, entity.users)]} )
        if entity.roles is not None and len(entity.roles) > 0 :        
            attrs.update( {ROLES : [(MOD_REPLACE, entity.roles)]} )
        if len(attrs) > 0:            
            conn = ldaphelper.open()                
            conn.modify_s(__get_dn(entity), attrs_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='Perm update failed, not found:' + entity.obj_name, id=global_ids.PERM_OP_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='Perm update failed result=' + str(e), id=global_ids.PERM_UPDATE_FAILED)
    except Exception as e:
        raise RbacError(msg='Perm update error=' + str(e), id=global_ids.PERM_UPDATE_FAILED)
    return entity


def delete ( entity ):
    try:
        conn = ldaphelper.open()        
        conn.delete_s(__get_dn(entity))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='Perm delete not found:' + entity.obj_name, id=global_ids.PERM_OP_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='Perm delete failed result=' + str(e), id=global_ids.PERM_DELETE_FAILED)
    except Exception as e:
        raise RbacError(msg='Perm delete error=' + str(e), id=global_ids.PERM_DELETE_FAILED)
    return entity


def grant ( entity, role ):
    try:
        attrs = {}
        # constraint type:
        if role is not None:
            attrs.update( {ROLES : [(MOD_ADD, role.name)]} )
            conn = ldaphelper.open()
            conn.modify_s(__get_dn(entity), attrs_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='Perm grant failed, not found, obj name='
                            +  entity.obj_name + ', op_name='
                            + entity.op_name
                            + ', op id=' + entity.obj_id
                            + ', role='+ role.name,
                            id=global_ids.PERM_OP_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='Perm grant failed result=' + str(e), id=global_ids.PERM_GRANT_FAILED)
    except Exception as e:
        raise RbacError(msg='Perm grant error=' + str(e), id=global_ids.PERM_GRANT_FAILED)
    return entity


def revoke ( entity, role ):
    try:
        attrs = {}
        # constraint type:
        if role is not None:
            attrs.update( {ROLES : [(MOD_DELETE, role.name)]} )
            conn = ldaphelper.open()
            conn.modify_s(__get_dn(entity), attrs_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise FortressError(msg='Perm not found, obj name='
                            +  entity.obj_name
                            + ', op_name='
                            + entity.op_name
                            + ', role='
                            + role.name,
                            id=global_ids.PERM_NOT_EXIST)
    except ldap.NO_SUCH_ATTRIBUTE:
        raise FortressError(msg='Perm revoke failed, not granted, obj name='
                            +  entity.obj_name
                            + ', op_name='
                            + entity.op_name
                            + ', role='
                            + role.name,
                            id=global_ids.PERM_ROLE_NOT_EXIST)
    except ldap.LDAPError as e:
        raise FortressError(msg='Perm revoke failed result=' + str(e), id=global_ids.PERM_REVOKE_FAILED)
    except Exception as e:
        raise FortressError(msg='Perm revoke error=' + str(e), id=global_ids.PERM_REVOKE_FAILED)
    return entity


def create_obj ( entity ):
    try:
        attrs = {}
        attrs.update( {'objectClass': PERM_OBJ_OCS} )
        attrs.update( {OBJ_NM : entity.obj_name} )
        # generate random id:
        entity.internal_id = str(uuid.uuid4())
        attrs.update( {global_ids.INTERNAL_ID : entity.internal_id} )
        # ou is req'd for organizationalUnit object class, if caller did not set, use obj name.
        if not entity.ou:
            entity.ou = entity.obj_name
        attrs.update( {global_ids.OU : entity.ou} )
        if entity.description:        
            attrs.update( {global_ids.DESC : entity.description} )
        if entity.type:        
            attrs.update( {TYPE : entity.type} )
        # list of comma delimited strings:
        if entity.props is not None and len(entity.props) > 0 :        
            attrs.update( {global_ids.PROPS : entity.props} )
        conn = ldaphelper.open()        
        conn.add_s(__get_obj_dn(entity), add_to_modlist(attrs))
    except ldap.ALREADY_EXISTS:
        raise NotUnique(msg='PermObj create failed, already exists:' + entity.uid, id=global_ids.PERM_ADD_FAILED)
    except ldap.LDAPError as e:
        raise FortressError(msg='PermObj create failed result=' + str(e), id=global_ids.PERM_ADD_FAILED)
    except Exception as e:
        raise FortressError(msg='PermObj create error=' + str(e), id=global_ids.PERM_ADD_FAILED)
    return entity


def update_obj ( entity ):
    conn = None
    try:
        attrs = {}        
        if entity.ou:        
            attrs.update( {global_ids.OU : [(MOD_REPLACE, [entity.ou])]} )
        if entity.description:        
            attrs.update( {global_ids.DESC : [(MOD_REPLACE, [entity.description])]} )
        if entity.type:        
            attrs.update( {TYPE : [(MOD_REPLACE, [entity.type])]} )
        # list of comma delimited strings:
        if entity.props is not None and len(entity.props) > 0 :
            attrs.update( {global_ids.PROPS : [(MOD_REPLACE, entity.props)]} )
        if len(attrs) > 0:            
            conn = ldaphelper.open()                
            conn.modify_s(__get_obj_dn(entity), attrs_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='PermObj update failed, not found:' + entity.obj_name, id=global_ids.PERM_OBJ_NOT_FOUND)
    except ldap.LDAPError as e:
        raise FortressError(msg='PermObj update failed result=' + str(e), id=global_ids.PERM_UPDATE_FAILED)
    except Exception as e:
        raise FortressError(msg='PermObj update error=' + str(e), id=global_ids.PERM_UPDATE_FAILED)
    return entity


def delete_obj ( entity ):
    try:
        conn = ldaphelper.open()        
        conn.delete_s(__get_obj_dn(entity))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='PermObj delete not found:' + entity.obj_name, id=global_ids.PERM_OBJ_NOT_FOUND)
    except ldap.LDAPError as e:
        raise FortressError(msg='PermObj delete failed result=' + str(e), id=global_ids.PERM_DELETE_FAILED)
    except Exception as e:
        raise FortressError(msg='PermObj delete error=' + str(e), id=global_ids.PERM_DELETE_FAILED)
    return entity


def __get_obj_dn(entity):
    return OBJ_NM + '=' + entity.obj_name + "," + __CONTAINER_DN


def __get_dn(entity):
    dn = ''
    if entity.obj_id is not None and len(entity.obj_id) > 0:
        dn = OBJ_ID + '=' + entity.obj_id + '+' + OP_NM + '=' + entity.op_name + ',' + __get_obj_dn(entity)
    else:
        dn = OP_NM + '=' + entity.op_name + ',' + __get_obj_dn(entity)
    return dn


def attrs_to_modlist(attrs):
    result = []
    for attr, mods in attrs.items():
        for op, vals in mods:
            if not isinstance(vals, (tuple, list)):
                vals = [vals]
            vals = [v.encode() if isinstance(v, str) else v for v in vals]
            result.append((op, attr, vals))
    return result


PERM_OC_NAME = 'ftOperation'
PERM_OCS = [PERM_OC_NAME, global_ids.PROP_OC_NAME]
PERM_OBJ_OC_NAME = 'ftObject'
PERM_OBJ_OCS = [PERM_OBJ_OC_NAME, global_ids.PROP_OC_NAME]

ROLES = 'ftRoles'
OBJ_ID = 'ftObjId'
OBJ_NM = 'ftObjNm'
OP_NM = 'ftOpNm'
PERM_NAME = 'ftPermName'
USERS = 'ftUsers'
TYPE = 'ftType'

SEARCH_ATTRS = [
    global_ids.INTERNAL_ID, OBJ_NM, OP_NM, PERM_NAME, OBJ_ID, ROLES,
     USERS, TYPE, global_ids.PROPS, global_ids.DESC
     ]

SEARCH_OBJ_ATTRS = [
    global_ids.INTERNAL_ID, OBJ_NM, TYPE, global_ids.PROPS, global_ids.DESC, global_ids.OU
     ]

__CONTAINER_DN = ldaphelper.get_container_dn(global_ids.PERM_OU)



'''
@copyright: 2022 - Symas Corporation
'''
import sys
import argparse
from rbac.model import PermObj, Perm, User, Role, Constraint
from rbac import review, admin
from rbac.cli.utils import print_user, print_role, print_ln, print_entity
from rbac.util.fortress_error import RbacError
from rbac.cli.utils import (
    load_entity, add_args, USER, ROLE, PERM, OBJECT, ADD, 
    UPDATE, DELETE, ASSIGN, DEASSIGN, READ, SEARCH, GRANT, REVOKE
    )

def process(args):
    result = False
    try:
        if args.entity == USER:
            result = process_user(args)
        elif args.entity == ROLE:
            result = process_role(args)        
        elif args.entity == PERM:
            result = process_perm(args)                
        elif args.entity == OBJECT:
            result = process_object(args)
        else:
            print('process failed, invalid entity=' + args.entity)        
        if result:
            print('success')
    except RbacError as e:
        print('RbacError id=' + str(e.id) +', ' + e.msg)                        
                        

def process_user(args):
    user = load_entity(User(), args)        
    print(args.entity + ' ' + args.operation)    
    if not args.uid:
        print("error --uid required for entity user")    
        return False
    elif args.operation == ADD:
        if not args.name:
            args.name = args.uid
        constraint = load_entity(Constraint(), args)
        user.constraint = constraint
        admin.add_user(user)
    elif args.operation == UPDATE:
        if args.name is not None:        
            constraint = load_entity(Constraint(), args)
            user.constraint = constraint
        admin.update_user(user)
    elif args.operation == DELETE:
        admin.delete_user(user)
    elif args.operation == ASSIGN:
        role_nm = args.role        
        print('role=' + role_nm)
        admin.assign(user, Role(name=role_nm))
    elif args.operation == DEASSIGN:
        role_nm = args.role        
        print('role name=' + role_nm)        
        admin.deassign(user, Role(name=role_nm))
    elif args.operation == READ:
        print_user(review.read_user(user), user.uid)
        pass
    elif args.operation == SEARCH:
        user.uid += '*'
        users = review.find_users(user)
        if len(users) > 0:
            for idx, usr in enumerate(users):
                print_user(usr, user.uid + ':' + str(idx))
        else:
            print_ln('No matching records found matching filter: ' + user.uid) 
    else:
        print('process_user failed, invalid operation=' + args.operation)
        return False
    return True                                

        
def process_role(args):
    role = load_entity(Role(), args)
    print(args.entity + ' ' + args.operation)        
    if not args.name:
        print("error --name required for entity role")    
        return False
    elif args.operation == ADD:
        constraint = load_entity(Constraint(), args)
        role.constraint = constraint
        admin.add_role(role)
    elif args.operation == UPDATE:
        constraint = load_entity(Constraint(), args)
        role.constraint = constraint
        admin.update_role(role)
    elif args.operation == DELETE:
        admin.delete_role(role)
    elif args.operation == READ:
        print_role(review.read_role(role), role.name)
        pass
    elif args.operation == SEARCH:
        role.name += '*'
        roles = review.find_roles(role)
        if len(roles) > 0:
            for idx, rle in enumerate(roles):
                print_role(rle, role.name + ':' + str(idx))
        else:
            print_ln('No matching records found matching filter: ' + role.name) 
    else:
        print('process_role failed, invalid operation=' + args.operation)
        return False
    return True                                
       
        
def process_object(args):
    pobject = load_entity (PermObj(), args) 
    print(args.entity + ' ' + args.operation)   
    if not args.obj_name:
        print("error --obj_name required for entity object")    
        return False
    elif args.operation == ADD:
        admin.add_object(pobject)
    elif args.operation == UPDATE:
        admin.update_object(pobject)
    elif args.operation == DELETE:
        admin.delete_object(pobject)
    elif args.operation == READ:
        print_entity(review.read_object(pobject), pobject.obj_name)
        pass
    elif args.operation == SEARCH:
        pobject.obj_name += '*'
        objs = review.find_objects(pobject)
        if len(objs) > 0:
            for idx, obj in enumerate(objs):
                print_entity(obj, pobject.obj_name + ':' + str(idx))
        else:
            print_ln('No matching records found matching filter: ' + pobject.obj_name) 
    else:
        print('process_object failed, invalid operation=' + args.operation)
        return False
    return True                                
        
        
def process_perm(args):
    perm = load_entity (Perm(), args) 
    print(args.entity + ' ' + args.operation)       
    if args.operation == ADD:
        admin.add_perm(perm)
    elif args.operation == UPDATE:
        admin.update_perm(perm)
    elif args.operation == DELETE:
        admin.delete_perm(perm)
    elif args.operation == GRANT:
        role_nm = args.role        
        print('role=' + role_nm)
        admin.grant(perm, Role(name=role_nm))
    elif args.operation == REVOKE:
        role_nm = args.role        
        print('role=' + role_nm)
        admin.revoke(perm, Role(name=role_nm))
    elif args.operation == READ:
        print_entity(review.read_perm(perm), perm.obj_name + '.' + perm.op_name)
        pass
    elif args.operation == SEARCH:
        role_nm = args.role
        userid = args.uid
        prms = []
        label = ''
        if userid:
            label = userid
            prms = review.user_perms(User(uid=userid))
        elif role_nm:
            label = role_nm
            prms = review.role_perms(Role(name=role_nm))
        else:        
            if perm.obj_name:            
                perm.obj_name += '*'
            else:
                perm.obj_name = '*'            
            if perm.op_name:
                perm.op_name += '*'
            else:
                perm.op_name = '*'
            label = perm.obj_name + '.' + perm.op_name            
            prms = review.find_perms(perm)
        if len(prms) > 0:
            for idx, prm in enumerate(prms):
                print_entity(prm, label + ':' + str(idx))
        else:
            print_ln('No matching records found matching filter: ' + label)                                 
    else:
        print('process_perm failed, invalid operation=' + args.operation)
        return False
    return True                                
        
        
def main(argv=None): # IGNORE:C0111
    '''Command line options.'''

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    parser = argparse.ArgumentParser(description='Process py-fortress admin and review commands.')
    parser.add_argument('entity', metavar='entity', choices=[USER, ROLE, PERM, OBJECT], help='entity name')
    parser.add_argument('operation', metavar='operand', choices=[ADD, UPDATE, DELETE, ASSIGN, DEASSIGN, GRANT, REVOKE, READ, SEARCH], help='operation name')
    parser.add_argument('-r', '--role', dest='role', help='role name')
    parser.add_argument('--phones', nargs="*", default=[])
    parser.add_argument('--mobiles', nargs="*", default=[])
    parser.add_argument('--emails', nargs="*", default=[])
    parser.add_argument('--props', nargs="*", default=[])
        
    add_args(parser, Role())
    add_args(parser, User())
    add_args(parser, Perm())    
    add_args(parser, PermObj())
    add_args(parser, Constraint())
    args = parser.parse_args()
    process(args)

if __name__ == "__main__":    
    sys.exit(main())


'''
@copyright: 2022 - Symas Corporation
'''

class Role:    
    "Fortress Role"
    def __init__(
            self,            
            name=None,
            internal_id=None,
            description=None,
            props=None,
            constraint=None,
            members=None,
            dn=None
            ):
        self.name=name
        self.internal_id=internal_id
        self.description=description
        self.props=props
        self.constraint=constraint
        self.dn=dn
        self.members=members


'''
@copyright: 2022 - Symas Corporation
'''

from . import Validator
from .global_ids import ACTV_FAILED_DAY, SUCCESS

class Day(Validator):
        
    def validate(self, constraint, now, session=None):
        if constraint.day_mask == 'all' or not constraint.day_mask or constraint.day_mask is None:
            return SUCCESS                
        if now.day_of_week in constraint.day_mask:
            return SUCCESS
        else:
            return ACTV_FAILED_DAY        
        
        


'''
@copyright: 2022 - Symas Corporation
'''

class User:    
    "Fortress User"

    def __init__(
            self,
            uid=None,
            password=None,
            ou=None,
            internal_id=None,
            constraint=None,            
            role_constraints=None,            
            roles=None,
            pw_policy=None,
            cn=None,
            sn=None,
            dn=None,
            description=None,
            display_name=None,
            employee_type=None,
            title=None,
            phones=None,
            mobiles=None,
            emails=None,
            reset = None,
            locked_time = None,
            system = None,
            props = None,
            department_number = None,
            l = None,
            physical_delivery_office_name = None,
            postal_code = None,
            room_number = None,                        
            ):
        self.uid = uid
        self.password = password
        self.ou = ou
        self.internal_id = internal_id
        self.constraint = constraint        
        self.role_constraints = role_constraints        
        self.roles = roles
        self.pw_policy = pw_policy        
        self.cn = cn       
        self.sn = sn
        self.dn = dn
        self.description = description
        self.display_name = display_name
        self.employee_type = employee_type
        self.title = title
        self.phones = phones
        self.mobiles = mobiles
        self.emails = emails
        self.reset = reset
        self.locked_time = locked_time
        self.system = system
        self.props = props        
        self.department_number = department_number
        self.l = l
        self.physical_delivery_office_name = physical_delivery_office_name
        self.postal_code = postal_code
        self.room_number = room_number


'''
@copyright: 2022 - Symas Corporation
'''
from argparse import ArgumentError

# entities =
USER = 'user'
ROLE = 'role'
PERM = 'perm'
OBJECT = 'object'
#operations =
ADD = 'add'
UPDATE = 'mod'
DELETE = 'del'
ASSIGN = 'assign'
DEASSIGN = 'deassign'
GRANT = 'grant'
REVOKE = 'revoke'
READ = 'read'
SEARCH = 'search'

AUTH = 'auth'
CHCK = 'check'
ROLES = 'roles'
PERMS = 'perms'
SHOW = 'show'
DROP = 'drop'

def load_entity (entity, args):
    for name in entity.__dict__:
        value = args.__dict__[name]
        if value:
            entity.__dict__[name] = value
            if name != 'password':
                print(name + '=' + str(value))
    return entity

def add_args (parser, entity):
    for name in entity.__dict__:
        try:
            parser.add_argument('--' + name)
        except ArgumentError as e:
            #ignore
            pass


def print_ln(ln, num=None):
    if num is None:
        num = 0
    tabs = ''
    for x in range(0, num):
        tabs += '\t'
    print(tabs + ln)
    #logger.debug((tabs + ln))


def print_entity(entity, label, indent=None):
    if indent is None:
        indent = 1
    print_ln(label, indent - 1)
    for name in entity.__dict__:
        if name != 'password':
            value = entity.__dict__[name]
            if value:
                print_ln("\t{0}: {1}".format(name, value), indent)


def print_user(entity, label):
    print_entity(entity, label, 1)
    if entity.constraint is not None:
        print_entity(entity.constraint, "User Constraint:", 2)

    if entity.role_constraints is not None:
        for idx, constraint in enumerate(entity.role_constraints):
            # print_constraint (constraint, "User-Role Constraint[" + str(idx+1) + "]:")
            if constraint.name:
                print_entity(constraint, "User-Role Constraint[" + str(idx + 1) + "]:", 2)
    print_ln("*************** " + label + " *******************")


def print_role(entity, label):
    print_entity(entity, label, 1)
    if entity.constraint is not None and entity.constraint.name:
        print_entity(entity.constraint, "Role Constraint:", 2)
    print_ln("*************** " + label + " *******************")



'''
@copyright: 2022 - Symas Corporation
'''

import ldap
from rbac.ldap import ldaphelper, NotFound, NotUnique
from rbac.ldap.ldaphelper import add_to_modlist
from rbac.util import Config, global_ids
from rbac.util.fortress_error import RbacError
from rbac.util import logger


def create_ou (name, desc=None):
    __validate(name) 
    try:
        attrs = {}
        attrs.update( {'objectClass': OU_OCS} )
        attrs.update( {OU_NAME : name} )
        if not desc:
            desc = 'py-fortress Container ' + name
        attrs.update( {global_ids.DESC : desc} )
        conn = ldaphelper.open()  
        conn.add_s(ldaphelper.get_container_dn(name), add_to_modlist(attrs))
    except Exception as e:
        raise RbacError(msg='OU create error=' + str(e), id=global_ids.CNTR_CREATE_FAILED)
    except ldap.ALREADY_EXISTS:
        raise NotUnique(msg='OU create failed, already exists:' + name, id=global_ids.CNTR_ALREADY_EXISTS)
    except ldap.LDAPError as e:
        raise RbacError(msg='OU create failed result=' + str(e), id=global_ids.CNTR_CREATE_FAILED)


def delete_ou ( name ):
    __validate(name)   
    try:
        conn = ldaphelper.open()       
        conn.delete_s(ldaphelper.get_container_dn(name))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='OU delete not found:' + name, id=global_ids.CNTR_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='OU delete failed result=' + str(e), id=global_ids.CNTR_DELETE_FAILED)
    except Exception as e:
        raise RbacError(msg='OU delete error=' + str(e), id=global_ids.CNTR_DELETE_FAILED)


def create_suffix ( name ):
    dn = DC_NAME + '=' + name + ',' + DC_NAME + '=com'
    try:
        attrs = {}
        attrs.update( {'objectClass': SUFX_OCS} )
        attrs.update( {DC_NAME : name} )
        attrs.update( {O : name} )
        conn = ldaphelper.open()        
        conn.add_s(dn, add_to_modlist(attrs))
    except Exception as e:
        raise NotUnique(msg='Suffix create failed, already exists:' + dn, id=global_ids.SUFX_ALREADY_EXISTS)
    except ldap.ALREADY_EXISTS:
        raise RbacError(msg='Suffix create failed, dn=' + dn + ', result=' + str(e), id=global_ids.SUFX_CREATE_FAILED)
    except Exception as e:
        raise RbacError(msg='Suffix create dn=' + dn + ', error=' + str(e), id=global_ids.SUFX_CREATE_FAILED)


def delete_suffix ():
    try:
        conn = ldaphelper.open()        
        conn.delete_s(__SUFX_DN)
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='Suffix delete not found dn=' + __SUFX_DN, id=global_ids.SUFX_NOT_EXIST)
    except ldap.LDAPError as e:
        raise RbacError(msg='Suffix delete failed, dn=' + __SUFX_DN + ', result=' + str(e), id=global_ids.SUFX_DELETE_FAILED)
    except Exception as e:
        raise RbacError(msg='Suffix delete failed, dn=' + __SUFX_DN + ', error=' + str(e), id=global_ids.SUFX_DELETE_FAILED)


def bootstrap ():
    suffix_nm = __get_rdn_name(__SUFX_DN)
    try:
        create_suffix ( suffix_nm )
    except NotUnique:
        logger.warn('create suffix failed, already present, name=' + suffix_nm)
    try:
        create_ou (global_ids.USER_OU)
    except NotUnique:
        logger.warn('create users container failed, already present')
    try:
        create_ou (global_ids.ROLE_OU)
    except NotUnique:
        logger.warn('create roles container failed, already present')
    try:
        create_ou (global_ids.PERM_OU)
    except NotUnique:
        logger.warn('create perms container failed, already present')    
 
 
def __validate(name):
    if not name:
        raise RbacError(msg='OU name null', id=global_ids.CNTR_NAME_NULL)

 
def __get_rdn_name(dn):
    name = None
    values = dn.split(',')        
    values = [ val.strip() for val in values ]
    if values[0] is not None:
        name=values[0]
    return name[3:]  
 
  
# suffix:
__SUFX_DN = Config.get(global_ids.DIT)[global_ids.SUFFIX]
DC_OC_NAME = 'dcObject'
ORG_OC_NAME = 'organization'
SUFX_OCS = [DC_OC_NAME, ORG_OC_NAME]
DC_NAME = 'dc'
O = 'o'

# orgunit:
OU_OC_NAME = 'organizationalUnit'
OU_OCS = [OU_OC_NAME]
OU_NAME = 'ou'



'''
@copyright: 2022 - Symas Corporation
'''
from rbac.model.user import User
from rbac.model.role import Role
from rbac.model.perm import Perm
from rbac.model.perm_object import PermObj
from rbac.model.constraint import Constraint
#from rbac import review
#from rbac import access, admin
from rbac.util import global_ids
from rbac.util.fortress_error import RbacError



'''
@copyright: 2022 - Symas Corporation
'''

import uuid
import ldap
from ldap import MOD_REPLACE, MOD_ADD, MOD_DELETE
from ldap.cidict import cidict as CIDict
from rbac.model import User, Constraint
from rbac.ldap import ldaphelper, NotFound, NotUnique, InvalidCredentials
from rbac.ldap.ldaphelper import add_to_modlist, mods_to_modlist
from rbac.util import global_ids
from rbac.util.fortress_error import RbacError

def read (entity):
    userList = search(entity)
    if userList is None or len(userList) == 0:
        raise NotFound(msg="User Read not found, uid=" + entity.uid, id=global_ids.USER_NOT_FOUND)
    elif len(userList) > 1:
        raise NotUnique(msg="User Read not unique, uid=" + entity.uid, id=global_ids.USER_READ_FAILED)
    else:
        return userList[0]


def authenticate (entity):
    conn = None
    result = False            
    try:        
        conn = ldaphelper.open_user(__get_dn(entity), entity.password)
        # result = conn.bind() # TODO: WTH?
    except Exception as e:
        raise RbacError(msg='User Authenticate error for uid=' + entity.uid + ', LDAP error=' + str(e), id=global_ids.USER_PW_CHK_FAILED)
    finally:
        if conn:        
            ldaphelper.close_user(conn)
    return True


def search (entity):
    conn = None            
    userList = []
    search_filter = '(&(objectClass=' + USER_OC_NAME + ')'
    if entity.uid:
        search_filter += '(' + global_ids.UID + '=' + entity.uid + ')'
    if entity.ou:
        search_filter += '(' + global_ids.OU + '=' + entity.ou + ')'
    search_filter += ')'           
    try:
        conn = ldaphelper.open()
        entries = conn.search_s(CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=SEARCH_ATTRS)
        for dn, attrs in entries:
            userList.append(__unload(dn, attrs))
    except Exception as e:
        raise RbacError(msg='User Search error=' + str(e))
    finally:
        if conn:        
            ldaphelper.close(conn)
    return userList


# assumes that roles contains at least one role name
def search_on_roles (roles):    
    conn = None            
    userList = []    
    search_filter = '(&(objectClass=' + USER_OC_NAME + ')'
    if len (roles) > 1:
        search_filter += '(|'
        end_filter = '))'
    else:
        end_filter = ')'
    for role in roles:
        search_filter += '(' + ROLES + '=' + role + ')'
    search_filter += end_filter                    
    try:
        conn = ldaphelper.open()
        entries = conn.search_s(CONTAINER_DN, scope=ldap.SCOPE_SUBTREE, filterstr=search_filter, attrlist=SEARCH_ATTRS)
        for dn, attrs in entries:
            userList.append(__unload(dn, attrs))
    except Exception as e:
        raise RbacError(msg='User Search Roles error=' + str(e), id=global_ids.URLE_SEARCH_FAILED)
    finally:
        if conn:        
            ldaphelper.close(conn)
    return userList


def __unload(dn, attrs):
    entity = User()
    entity.dn = dn

    attrs = CIDict(attrs)

    entity.uid = ldaphelper.get_one_attr_val(attrs.get(global_ids.UID, []))
    entity.ou = ldaphelper.get_one_attr_val(attrs.get(global_ids.OU, []))
    entity.internal_id = ldaphelper.get_attr_val(attrs.get(global_ids.INTERNAL_ID, []))
    entity.pw_policy = ldaphelper.get_attr_val(attrs.get(PW_POLICY, []))
    entity.cn = ldaphelper.get_one_attr_val(attrs.get(global_ids.CN, []))
    entity.sn = ldaphelper.get_one_attr_val(attrs.get(global_ids.SN, []))
    entity.description = ldaphelper.get_one_attr_val(attrs.get(global_ids.DESC, []))
    entity.display_name = ldaphelper.get_attr_val(attrs.get(DISPLAY_NAME, []))
    entity.employee_type = ldaphelper.get_one_attr_val(attrs.get(EMPLOYEE_TYPE, []))
    entity.title = ldaphelper.get_one_attr_val(attrs.get(TITLE, []))
    entity.reset = ldaphelper.get_bool(attrs.get(IS_RESET, []))
    entity.system = ldaphelper.get_bool(attrs.get(IS_SYSTEM, []))
    entity.department_number = ldaphelper.get_one_attr_val(attrs.get(DEPT_NUM, []))
    entity.l = ldaphelper.get_one_attr_val(attrs.get(LOCATION, []))
    entity.physical_delivery_office_name = ldaphelper.get_one_attr_val(attrs.get(PHYSICAL_OFFICE_NM, []))
    entity.postal_code = ldaphelper.get_one_attr_val(attrs.get(POSTAL_CODE, []))
    entity.room_number = ldaphelper.get_one_attr_val(attrs.get(RM_NUM, []))

    # Get the attr as object:
    entity.locked_time = ldaphelper.get_attr_object(attrs.get(LOCKED_TIME, []))

    # Get the multi-occurring attrs:
    entity.props = ldaphelper.get_list(attrs.get(global_ids.PROPS, []))
    entity.phones = ldaphelper.get_list(attrs.get(TELEPHONE_NUMBER, []))
    entity.mobiles = ldaphelper.get_list(attrs.get(MOBILE, []))
    entity.emails = ldaphelper.get_list(attrs.get(MAIL, []))
    entity.roles = ldaphelper.get_list(attrs.get(ROLES, []))

    # unload raw user constraint:
    entity.constraint = Constraint(ldaphelper.get_attr_val(attrs.get(global_ids.CONSTRAINT, [])))

    # now, unload raw user-role constraints:
    rcsRaw = ldaphelper.get_list(attrs.get(ROLE_CONSTRAINTS, []))
    if rcsRaw is not None :
        entity.role_constraints = []
        for rcRaw in rcsRaw :
            entity.role_constraints.append(Constraint(rcRaw))
                        
    return entity


def create ( entity ):
    try:
        attrs = {}
        attrs.update( {'objectClass': USER_OCS} )
        attrs.update( {global_ids.UID : entity.uid} )
        # generate random id:
        entity.internal_id = str(uuid.uuid4())
        attrs.update( {global_ids.INTERNAL_ID : entity.internal_id} )        
        # cn is req'd for iNetOrgPerson, if caller did not set, use uid value
        if not entity.cn:
            entity.cn = entity.uid
        attrs.update( {global_ids.CN : entity.cn} )
        # likewise sn is req'd for iNetOrgPerson, if caller did not set, use uid value
        if not entity.sn:
            entity.sn = entity.uid
        attrs.update( {global_ids.SN : entity.sn} )  
        # strings:  
        if entity.password:                
            attrs.update( {PW : entity.password} )
        if entity.description:        
            attrs.update( {global_ids.DESC : entity.description} )
        if entity.ou :        
            attrs.update( {global_ids.OU : entity.ou} )
        if entity.display_name:        
            attrs.update( {DISPLAY_NAME : entity.display_name} )
        if entity.employee_type:        
            attrs.update( {EMPLOYEE_TYPE : entity.employee_type} )
        if entity.title:        
            attrs.update( {TITLE : entity.title} )
        if entity.department_number:        
            attrs.update( {DEPT_NUM : entity.department_number} )
        if entity.l:        
            attrs.update( {LOCATION : entity.l} )
        if entity.physical_delivery_office_name:        
            attrs.update( {PHYSICAL_OFFICE_NM : entity.physical_delivery_office_name} )
        if entity.postal_code:        
            attrs.update( {POSTAL_CODE : entity.postal_code} )
        if entity.room_number:        
            attrs.update( {RM_NUM : entity.room_number} )                        
        if entity.pw_policy:        
            attrs.update( {PW_POLICY : entity.pw_policy} )
        # boolean:
        if entity.system is not None :        
            attrs.update( {IS_SYSTEM : 'TRUE' if entity.system else 'FALSE'} )
        # list of strings:
        if entity.phones is not None and len(entity.phones) > 0 :        
            attrs.update( {TELEPHONE_NUMBER : entity.phones} )
        if entity.mobiles is not None and len(entity.mobiles) > 0 :        
            attrs.update( {MOBILE : entity.mobiles} )
        if entity.emails is not None and len(entity.emails) > 0 :        
            attrs.update( {MAIL : entity.emails} )            
        if entity.props is not None and len(entity.props) > 0 :        
            attrs.update( {global_ids.PROPS : entity.props} )
        # list of delimited strings:
        if entity.constraint:        
            attrs.update( {global_ids.CONSTRAINT : entity.constraint.get_raw()} )
            
        conn = ldaphelper.open()     
        conn.add_s(__get_dn(entity), add_to_modlist(attrs))
    except ldap.ALREADY_EXISTS:
        raise NotUnique(msg='User create failed, already exists:' + entity.uid, id=global_ids.USER_ADD_FAILED)
    except ldap.LDAPError as e:
        raise RbacError(msg='User create failed result=' + str(e), id=global_ids.USER_ADD_FAILED)
    except Exception as e:
        raise RbacError(msg='User create error=' + str(e), id=global_ids.USER_ADD_FAILED)
    return entity


def update ( entity ):
    try:
        attrs = {}
        # strings:                
        if entity.cn:
            attrs.update( {global_ids.CN : [(MOD_REPLACE, [entity.cn])]} )
        if entity.sn:
            attrs.update( {global_ids.SN : [(MOD_REPLACE, [entity.sn])]} )
        if entity.password:                
            attrs.update( {PW : [(MOD_REPLACE, [entity.password])]} )
        if entity.description:        
            attrs.update( {global_ids.DESC : [(MOD_REPLACE, [entity.description])]} )
        if entity.ou:        
            attrs.update( {global_ids.OU : [(MOD_REPLACE, [entity.ou])]} )
        if entity.display_name:        
            attrs.update( {DISPLAY_NAME : [(MOD_REPLACE, [entity.display_name])]} )
        if entity.employee_type:        
            attrs.update( {EMPLOYEE_TYPE : [(MOD_REPLACE, entity.employee_type)]} )
        if entity.title:        
            attrs.update( {TITLE : [(MOD_REPLACE, [entity.title])]} )
        if entity.department_number:        
            attrs.update( {DEPT_NUM : [(MOD_REPLACE, entity.department_number)]} )
        if entity.l:        
            attrs.update( {LOCATION : [(MOD_REPLACE, entity.l)]} )
        if entity.physical_delivery_office_name:        
            attrs.update( {PHYSICAL_OFFICE_NM : [(MOD_REPLACE, entity.physical_delivery_office_name)]} )
        if entity.postal_code:        
            attrs.update( {POSTAL_CODE : [(MOD_REPLACE, entity.postal_code)]} )
        if entity.room_number:        
            attrs.update( {RM_NUM : [(MOD_REPLACE, entity.room_number)]} )
        if entity.pw_policy:        
            attrs.update( {PW_POLICY : [(MOD_REPLACE, entity.pw_policy)]} )
            
        # list of strings:
        if entity.phones is not None and len(entity.phones) > 0 :        
            attrs.update( {TELEPHONE_NUMBER : [(MOD_REPLACE, entity.phones)]} )
        if entity.mobiles is not None and len(entity.mobiles) > 0 :        
            attrs.update( {MOBILE : [(MOD_REPLACE, entity.mobiles)]} )
        if entity.emails is not None and len(entity.emails) > 0 :        
            attrs.update( {MAIL : [(MOD_REPLACE, entity.emails)]} )
        if entity.system is not None :        
            attrs.update( {IS_SYSTEM : [(MOD_REPLACE, 'TRUE' if entity.system else 'FALSE')]} )

        # list of delimited strings::
        if entity.constraint is not None :        
            attrs.update( {global_ids.CONSTRAINT : [(MOD_REPLACE, entity.constraint.get_raw())]} )
            
        # boolean:
        if entity.props is not None and len(entity.props) > 0 :        
            attrs.update( {global_ids.PROPS : [(MOD_REPLACE, entity.props)]} )
            
        if len(attrs) > 0:            
            conn = ldaphelper.open()                
            conn.modify_s(__get_dn(entity), mods_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise NotFound(msg='User update failed, not found:' + entity.name, id=global_ids.USER_UPDATE_FAILED)
    except ldap.LDAPError as e:
        raise RbacError(msg='User update failed result=' + str(e), id=global_ids.USER_UPDATE_FAILED)
    except Exception as e:
        raise RbacError(msg='User update error=' + str(e), id=global_ids.USER_UPDATE_FAILED)
    return entity


def delete ( entity ):
    try:
        conn = ldaphelper.open()        
        conn.delete_s(__get_dn(entity))
    except ldap.NO_SUCH_OBJECT:
        raise RbacError(msg='User delete not found:' + entity.uid, id=global_ids.USER_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='User delete failed result=' + str(e), id=global_ids.USER_DELETE_FAILED)
    except Exception as e:
        raise RbacError(msg='User delete error=' + str(e), id=global_ids.USER_DELETE_FAILED)
    return entity


def assign ( entity, constraint ):
    try:
        attrs = {}
        if constraint is not None:
            attrs.update( {ROLE_CONSTRAINTS : [(MOD_ADD, constraint.get_raw())]} )
            attrs.update( {ROLES : [(MOD_ADD, constraint.name)]} )
        if len(attrs) > 0:            
            conn = ldaphelper.open()                
            conn.modify_s(__get_dn(entity), mods_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise RbacError(msg='User assign failed, not found:' + entity.uid, id=global_ids.USER_NOT_FOUND)
    except ldap.LDAPError as e:
        raise RbacError(msg='User assign failed result=' + str(e), id=global_ids.URLE_ASSIGN_FAILED)
    except Exception as e:
        raise RbacError(msg='User assign error=' + str(e), id=global_ids.URLE_ASSIGN_FAILED)
    return entity


def deassign ( entity, constraint ):
    try:
        attrs = {}
        if constraint is not None:
            attrs.update( {ROLE_CONSTRAINTS : [(MOD_DELETE, constraint.get_raw())]} )
            attrs.update( {ROLES : [(MOD_DELETE, constraint.name)]} )
        if len(attrs) > 0:            
            conn = ldaphelper.open()                
            conn.modify_s(__get_dn(entity), mods_to_modlist(attrs))
    except ldap.NO_SUCH_OBJECT:
        raise RbacError(msg='User deassign failed, not found:' + entity.uid, id=global_ids.USER_NOT_FOUND)
    except ldap.NO_SUCH_ATTRIBUTE:
        raise RbacError(msg='User deassign failed, no such attribute=' + constraint.name, id=global_ids.URLE_ASSIGN_NOT_EXIST)
    except ldap.LDAPError as e:
        raise RbacError(msg='User deassign failed result=' + str(e), id=global_ids.URLE_ASSIGN_FAILED)
    except Exception as e:
        raise RbacError(msg='User deassign error=' + str(e), id=global_ids.URLE_DEASSIGN_FAILED)
    return entity


def __get_dn(entity):
    return global_ids.UID + '=' + entity.uid + ',' + CONTAINER_DN


USER_OC_NAME = 'inetOrgPerson'
USER_ATTRS_OC_NAME = 'ftUserAttrs'
USER_OCS = [USER_OC_NAME, USER_ATTRS_OC_NAME, global_ids.PROP_OC_NAME]

PW = 'userPassword'
ROLES = 'ftra'
PW_POLICY = 'pwdPolicySubentry'
ROLE_CONSTRAINTS = 'ftRC'
DISPLAY_NAME = 'displayName'
EMPLOYEE_TYPE = 'employeeType'
TITLE = 'title'
TELEPHONE_NUMBER = 'telephoneNumber'
MOBILE = 'mobile'
MAIL = 'mail'
IS_RESET = 'pwdReset'
LOCKED_TIME = 'pwdAccountLockedTime'
IS_SYSTEM = 'ftSystem'
DEPT_NUM = 'departmentNumber'
LOCATION = 'l'
PHYSICAL_OFFICE_NM = 'physicalDeliveryOfficeName'
POSTAL_CODE = 'postalCode'
RM_NUM = 'roomNumber'

SEARCH_ATTRS = [
    global_ids.UID, global_ids.OU, global_ids.INTERNAL_ID, ROLES, ROLE_CONSTRAINTS, PW_POLICY, global_ids.CONSTRAINT,
    global_ids.CN, global_ids.SN, global_ids.DESC, DISPLAY_NAME, EMPLOYEE_TYPE,
    TITLE, TELEPHONE_NUMBER, MOBILE, MAIL, IS_RESET,
    LOCKED_TIME, IS_SYSTEM, global_ids.PROPS, DEPT_NUM,
    PHYSICAL_OFFICE_NM, POSTAL_CODE, RM_NUM, LOCATION
    ]

CONTAINER_DN = ldaphelper.get_container_dn(global_ids.USER_OU)



'''
@copyright: 2022 - Symas Corporation
'''

class Perm:    
    "Fortress Perm"
    def __init__(
            self,
            obj_name=None,
            op_name=None,
            obj_id=None,
            description=None,
            abstract_name=None,
            internal_id=None,
            type=None,
            users=None,
            roles=None,
            props=None,
            dn=None
            ):
        self.obj_name=obj_name
        self.op_name=op_name
        self.obj_id=obj_id
        self.description=description
        self.abstract_name=abstract_name
        self.internal_id=internal_id
        self.type=type
        self.users=users
        self.roles=roles
        self.props=props
        self.dn=dn


'''
@copyright: 2022 - Symas Corporation
'''

from .config import Config
from .validator import Validator
from .date import Date
from .day import Day
from .lockdate import LockDate
from .time import Time
from .timeout import TimeOut
from .current_date_time import CurrentDateTime
from .global_ids import ACTV_FAILED_DAY, ACTV_FAILED_DATE, ACTV_FAILED_TIMEOUT, ACTV_FAILED_TIME, ACTV_FAILED_LOCK, SUCCESS
from .fortress_error import RbacError
from .logger import logger


'''
@copyright: 2022 - Symas Corporation
'''

from . import Validator
from .global_ids import ACTV_FAILED_LOCK, SUCCESS

class LockDate(Validator):
        
    def validate(self, constraint, now, session=None):
        if not constraint.begin_lock_date or constraint.begin_lock_date is None or not constraint.end_lock_date or constraint.end_lock_date is None:
            return SUCCESS                
        elif constraint.begin_lock_date == 'none' or constraint.end_lock_date == 'none':
            return SUCCESS                
        elif constraint.begin_lock_date <= now.date <= constraint.end_lock_date:
            return ACTV_FAILED_LOCK
        else:
            return SUCCESS        
        
        


'''
@copyright: 2022 - Symas Corporation
'''
class PermObj:    
    "Fortress Permission Object"
    def __init__(
            self,
            obj_name=None,
            description=None,
            internal_id=None,
            ou=None,
            type=None,
            props=None,
            dn=None
            ):
        self.obj_name=obj_name
        self.description=description
        self.internal_id=internal_id
        self.ou=ou
        self.type=type
        self.props=props
        self.dn=dn


'''
@copyright: 2022 - Symas Corporation
'''

import json
import os
from . fortress_error import RbacError
from . import global_ids

# config file ENV VAR:
PYFORTRESS_CONF = "PYFORTRESS_CONF"
DATA = 'data'
FILENAME = 'filename'

class Config:
    current = {
        DATA: {},
        FILENAME: None
    }
    
    def load(filename='py-fortress-cfg.json'):
        found = False
        conf_folder = os.getenv(PYFORTRESS_CONF)
        if conf_folder is None:
            conf_folder=''
        for loc in os.curdir, os.path.expanduser("~"), "/etc/pyfortress", conf_folder:
            file = os.path.join(loc, filename)
            if os.path.isfile(file):
                with open(os.path.join(loc, filename)) as json_file:
                    #print("opening config file: " + file)
                    Config.current[DATA] = json.load(json_file)
                    Config.current[FILENAME] = filename
                    found = True
                    break
        if not found:
            msg = "Could not locate py-fortress-cfg.json. Was it added to current directory or user home directory or /etc/pyfortress or env var: " + PYFORTRESS_CONF
            print(msg)
            raise RbacError(msg="Configuration error=" + msg, id=global_ids.CONFIG_BOOTSTRAP_FAILED)

    def get(key):
        return Config.current[DATA][key]

    def getDefault(key, default=None):
        return Config.current[DATA].get(key, default)

# bootstrap the config:
Config.load()



'''
@copyright: 2022 - Symas Corporation
'''

class Session:    
    "Fortress Session"
    def __init__(
            self,
            user=None,
            is_authenticated=False,
            session_id=None,
            last_access=None,
            timeout=None,
            error_id=None,
            expiration_seconds=None,
            grace_logins=None,
            message=None,
            warnings=None
            ):
        self.user=user
        self.is_authenticated=is_authenticated
        self.session_id=session_id
        self.last_access=last_access
        self.timeout=timeout
        self.error_id=error_id
        self.expiration_seconds=expiration_seconds
        self.grace_logins=grace_logins
        self.message=message
        self.warnings=warnings


'''
@copyright: 2022 - Symas Corporation
'''

from rbac.model import Role, Constraint

ALL = 'all'
NONE = 'none'
BEGIN_DATE = '20180101'
END_DATE = '20501010'
BEGIN_LOCK_DATE = NONE
END_LOCK_DATE = NONE
BEGIN_TIME = '0000'
END_TIME = '0000'
TIMEOUT = 15

def get_test_roles(name, total):
    rles = []
    for i in range(total):
        rles.append(__create_test_role( name, i))
    return rles

    
def __create_test_role(name, number):
    
    label = name + '-' + str(number)
    return Role(
        name=label, 
        description=label + ' Role',
        constraint=Constraint(
            name=label,
            timeout=TIMEOUT,
            begin_time=BEGIN_TIME,
            end_time=END_TIME,
            begin_date=BEGIN_DATE,
            end_date=END_DATE,
            begin_lock_date=BEGIN_LOCK_DATE,
            end_lock_date=END_LOCK_DATE,
            day_mask=ALL
        )
    )


'''
@copyright:   2022 - Symas Corporation
'''
import sys
import pickle
import argparse
from rbac.util import global_ids
from rbac.model import Perm, User
from rbac import access
from rbac.util import RbacError
from ..cli.utils import print_user, print_entity
from rbac.cli.utils import (
    load_entity, add_args, ADD, DELETE, AUTH, CHCK, ROLES, PERMS, SHOW, DROP
    )

OUT_SESS_FILE = "sess.pickle"

def process(args):
    sess = None
    result = False
    user = load_entity (User(), args)
    perm = load_entity (Perm(), args)
    print(args.operation)        
    try:
        if args.operation == AUTH:
            sess = access.create_session(user, False)
            result = True
        elif args.operation == CHCK:
            sess = un_pickle()
            result = access.check_access(sess, perm)
        elif args.operation == ROLES:   
            sess = un_pickle()         
            roles = access.session_roles(sess)
            for idx, role in enumerate(roles):
                print_entity(role, role.name + ':' + str(idx))
            result = True
        elif args.operation == PERMS:   
            sess = un_pickle()         
            perms = access.session_perms(sess)
            for idx, perm in enumerate(perms):
                print_entity(perm, perm.obj_name + '.' + perm.op_name + ':' + str(idx))
            result = True
        elif args.operation == SHOW:     
            sess = un_pickle()       
            print_entity(sess, 'session')
            print_user(sess.user, 'user')           
            result = True
        elif args.operation == ADD:
            sess = un_pickle()
            if not args.role:
                print("error --role required for this op")    
                return False            
            print('role=' + args.role)
            access.add_active_role(sess, args.role)
            result = True
        elif args.operation == DROP:
            sess = un_pickle()
            if not args.role:
                print("error --role required for this op")    
                return False            
            print('role=' + args.role)            
            access.drop_active_role(sess, args.role)
            result = True
        else:
            print('process failed, invalid operation=' + args.operation)        
        if result:
            print('success')
        else: 
            print('failed')
        pickle_it(sess)
            
    except RbacError as e:
        if e.id == global_ids.ACTV_FAILED_DAY:
            print('failed day of week, id=' + str(e.id) + ', msg=' + e.msg)
        elif e.id == global_ids.ACTV_FAILED_DATE:
            print('failed for date, id=' + str(e.id) + ', msg=' + e.msg)
        elif e.id == global_ids.ACTV_FAILED_TIME:
            print('failed for time of day, id=' + str(e.id) + ', msg=' + e.msg)
        elif e.id == global_ids.ACTV_FAILED_TIMEOUT:
            print('failed inactivity timeout, id=' + str(e.id) + ', msg=' + e.msg)
        elif e.id == global_ids.ACTV_FAILED_LOCK:
            print('failed locked date')
        else:
            print('RbacError id=' + str(e.id) +', ' + e.msg)

                        
def pickle_it(sess):
    if sess is not None:
        pickling_on = open(OUT_SESS_FILE,"wb")
        pickle.dump(sess, pickling_on)
        pickling_on.close()


def un_pickle():
    pickle_off = open(OUT_SESS_FILE,"rb")
    sess = pickle.load(pickle_off)
    return sess


def main(argv=None):
    '''Command line options.'''
    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    program_name = 'Process py-fortress access commands.'
    parser = argparse.ArgumentParser(description=program_name)        
    parser.add_argument('operation', metavar='operand', choices=[AUTH,CHCK,ROLES,PERMS,ADD,DELETE,SHOW,DROP], help='operation name')
    parser.add_argument('-r', '--role', dest='role', help='role name')
    add_args(parser, User())
    add_args(parser, Perm())    
    args = parser.parse_args()
    process(args)


if __name__ == "__main__":    
    sys.exit(main())


'''
@copyright: 2022 - Symas Corporation
'''
from .user import User
from .constraint import Constraint
from .perm import Perm
from .perm_object import PermObj
from .role import Role
from .session import Session


'''
@copyright: 2022 - Symas Corporation
'''

import unittest
from rbac.ldap import InvalidCredentials
from rbac import access, review
from rbac.model import User
from rbac.cli.utils import print_ln
import sys


class BasicTestSuite(unittest.TestCase):
    """These tests the py-fortress access functions."""
                

class TestAccessMgr(unittest.TestCase):
    """
    Test the access funcs
    """    

    def test01_create_sessions(self):
        """
        Test the create_session method
        """
        print_ln('test create sessions')
                                             
        try:
            usr = User(uid = "py-user*")    
            uList = review.find_users(usr)
            loop_cnt = len(uList)
            for idx, entity in enumerate(uList) :
                if idx % 10 == 0:
                    sys.stdout.write('')
                    sys.stdout.flush()
                entity.password = 'password'
                try:
                    session = access.create_session(entity, False)
                    if session is None:
                        self.fail('test create sessions failed ' + entity.uid)
                except InvalidCredentials as e:
                    print_ln(e.msg)                    
            print()
        except Exception as e:
            self.fail('user create_session exception=' + e.msg)


    def test02_user_roles(self):
        """
        Test the user_roles & is_user_in_role method
        """
        print_ln('test user_roles')        
        try:
            usr = User(uid = "py-user*")
            uList = review.find_users(usr)
            for idx, entity in enumerate(uList) :
                if idx % 10 == 0:
                    sys.stdout.write('')
                    sys.stdout.flush()           
                entity.password = 'password'     
                session = access.create_session(entity, False)
                if session is None:
                    self.fail('test_user_roles failed ' + entity.uid)
                    
                if session.user.roles is not None and len(session.user.roles) > 0:                    
                    roles = access.session_roles(session)
                    for role in roles:
                        result = access.is_user_in_role(session, role.name)
                        if not result:
                            self.fail('test_user_roles failed uid=' + entity.uid + ', role=' + role)
            print()                                                                            
        except Exception as e:
            self.fail('test_user_roles exception=' + e)


    def test03_active_roles(self):
        """
        Test the add_active_role & drop_active_role methods
        """
        print_ln('test active_roles')        
        try:
            usr = User(uid = "py-user*")
            uList = review.find_users(usr)
            for idx, entity in enumerate(uList) :
                if idx % 10 == 0:
                    sys.stdout.write('')
                    sys.stdout.flush()                    
                
                entity.password = 'password'
                session = access.create_session(entity, False)
                if session is None:
                    self.fail('test_active_roles failed ' + entity.uid)
                    
                if session.user.roles is not None and len(session.user.roles) > 0:                    
                    roles = access.session_roles(session)
                    active_roles = list(roles)
                    
                    # now deactivate all of the roles:
                    for idx2, role in enumerate(active_roles):
                        if idx2 % 10 == 0:
                            sys.stdout.write('`')
                            sys.stdout.flush()                            
                        
                        result = access.is_user_in_role(session, role.name)
                        if not result:
                            self.fail('test_active_roles failed uid=' + entity.uid + ', role=' + role)
                        access.drop_active_role(session, role.name)
                        result = access.is_user_in_role(session, role.name)
                        if result:
                            self.fail('test_active_roles negative failed uid=' + entity.uid + ', role=' + role)
                            
                    if session.user.roles is None or len(session.user.roles) > 0:                    
                        self.fail('test_active_roles did not inactivate all roles in session for uid=' + entity.uid)
                            
                    # now activate all of the roles once again:
                    for idx2, role in enumerate(active_roles):
                        if idx2 % 10 == 0:
                            sys.stdout.write('`')
                            sys.stdout.flush()
                            
                        result = access.is_user_in_role(session, role.name)
                        if result:
                            self.fail('test_active_roles failed inactive negative check uid=' + entity.uid + ', role=' + role)
                        access.add_active_role(session, role.name)
                        result = access.is_user_in_role(session, role.name)
                        if not result:
                            self.fail('test_active_roles failed inactive positive check uid=' + entity.uid + ', role=' + role)
                            
                    if session.user.roles is None or len(session.user.roles) != len(active_roles):                    
                        self.fail('test_active_roles did not activate all roles in session for uid=' + entity.uid)
                            
            print()                                                                            
        except Exception as e:
            self.fail('test_active_roles exception=' + e.msg)


    def test04_session_permissions(self):
        """
        Test the session_perms & check_access method
        """
        print_ln('test session_perms')        
        try:
            usr = User(uid = "py-user*")
            uList = review.find_users(usr)
            for idx, entity in enumerate(uList) :
                if idx % 10 == 0:
                    sys.stdout.write('')
                    sys.stdout.flush()

                entity.password = 'password'                                    
                session = access.create_session(entity, False)
                if session is None:
                    self.fail('test_session_permissions failed ' + entity.uid)
                if session.user.roles is not None and len(session.user.roles) > 0:                    
                    perms = access.session_perms(session)
                    for idx2, perm in enumerate(perms):
                        if idx2 % 10 == 0:
                            sys.stdout.write('`')
                            sys.stdout.flush()
                                                        
                        result = access.check_access(session, perm)
                        if not result:
                            self.fail('test_session_permissions failed uid=' + entity.uid + ', perm obj name=' + perm.obj_name + ', op name=' + perm.op_name + ', obj id=' + perm.obj_id)                                                
            print()
        except Exception as e:
            self.fail('test_session_permissions exception=' + e.msg)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestAccessMgr('test01_create_sessions'))
    suite.addTest(TestAccessMgr('test02_user_roles'))
    suite.addTest(TestAccessMgr('test03_active_roles'))
    suite.addTest(TestAccessMgr('test04_session_permissions'))
       
    return suite  

 
if __name__ == '__main__':
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())


'''
@copyright: 2022 - Symas Corporation
'''
from ldap import set_option, LDAPError, OPT_X_TLS_REQUIRE_CERT, OPT_DEBUG_LEVEL, OPT_X_TLS_NEVER
from rbac.util import global_ids
from rbac.ldap import LdapException
from rbac.util import Config
from rbac.util import logger
from ldappool import ConnectionManager


# Open a connection for a particular user:
def open_user (user_dn, password):
    c = _open_user(user_dn, password)    
    return c

# Open a connection from the pool with service accounts:
def open ():
    c = _open_admin()
    return c

def _open_user (user_dn, password):
    try:
        with _srv_pool.connection(user_dn, password) as conn:
            c = conn
    except LDAPError as l:
        raise LdapException ('connutl.open user LDAPError=' + str (l))
    except Exception as e:
        raise LdapException ('connutl.open user Exception=' + str (e))
    return c

def _open_admin ():
    try:
        with _srv_pool.connection() as conn:
            c = conn
    except LDAPError as l:
        raise LdapException ('connutl.open admin LDAPError=' + str (l))
    except Exception as e:
        raise LdapException ('connutl.open admin Exception=' + str (e))
    return c

def close (conn):
    pass
    #conn.unbind()

def close_user (conn):
    conn.unbind()

def get_response(conn, id):
    res = conn.get_response(id)
    return res[0] 

# Call this to get value from a single-occurring attribute:    
def get_attr_val(lattr):
    value = ""
    if lattr:
        value = lattr[0].decode()
    return value

# Call this to get value from a single-occurring attribute:    
def get_attr_object(lattr):
    value = ""
    if lattr:
        value = lattr[0].decode()
    return value

# Call this when expecting a single value from a multi-occurring attribute:
def get_one_attr_val(lattr):
    if not lattr:
        return "" # FIXME: really?
    return lattr[0].decode()

# Call this to get a list of attribute values:
def get_list(lattr):
    return [value.decode() for value in lattr]

def get_bool(lattr, default=False):
    if not lattr:
        return default
    return (lattr[0].decode() == 'TRUE')

def get_result(conn, id):
    res = conn.get_response(id)
    return res[1].get('result')

def get_dn(entry):
    return entry['dn']

def get_container_dn(ou):
    return global_ids.OU + '=' + Config.get('dit')[ou] + ',' + __SUFX_DN

def add_to_modlist(attrs):
    result = []
    for attr, vals in attrs.items():
        if not isinstance(vals, (tuple, list)):
            vals = [vals]
        vals = [v.encode() if isinstance(v, str) else v for v in vals]
        result.append((attr, vals))
    return result

def mods_to_modlist(attrs):
    result = []
    for attr, mods in attrs.items():
        for op, vals in mods:
            if not isinstance(vals, (tuple, list)):
                vals = [vals]
            vals = [v.encode() if isinstance(v, str) else v for v in vals]
            result.append((op, attr, vals))
    return result


# Begin the Config section:
Config.load('py-fortress-cfg.json')
LDAP = 'ldap'
_service_uid = Config.get(LDAP)['dn']
_service_pw = Config.get(LDAP)['password']
_ldap_timeout = int(Config.get(LDAP)['timeout'])
_pool_size = int(Config.get(LDAP)['pool_size'])
_ldap_use_tls = Config.get(LDAP)['use_tls']
_uri = Config.get(LDAP)['uri']
_ldap_debug = Config.get(LDAP)['debug']
_ldap_validate_cacert = Config.get(LDAP)['validate_cacert']

_srv_pool = ConnectionManager(_uri, size=_pool_size, bind=_service_uid, passwd=_service_pw, timeout=_ldap_timeout, use_tls=_ldap_use_tls)
_usr_pool = ConnectionManager(_uri, size=_pool_size, timeout=_ldap_timeout, use_tls=_ldap_use_tls)
__SUFX_DN = Config.get('dit')['suffix']

if _ldap_validate_cacert == False:
    set_option(OPT_X_TLS_REQUIRE_CERT, OPT_X_TLS_NEVER)

if _ldap_debug:
    set_option(OPT_DEBUG_LEVEL, 255)



'''
@copyright: 2022 - Symas Corporation
'''
from rbac.ldap import permdao, userdao, roledao
from rbac.util import utils


def read_user(user):
    """
    Method returns matching User entity that is contained within the people container in the directory. 
    
    required parameters:
    user.uid - maps to INetOrgPerson uid  
    
    return:
    User   
    """    
    utils.validate_user(user)
    return userdao.read(user)

    
def find_users(user):
    """
    Return a list of type User of all users in the people container that match all or part of the User.userId field passed in User entity. 
    
    required parameters:
    user.uid - maps to existing user, can be partial with wildcard on end - *
    
    optional parameters:
    user.ou - maps to attribute assignment, can be partial with wildcard on end - *.
         
    return:
    User list   
    """    
    utils.validate_user(user)
    return userdao.search(user)

    
def read_role(role):
    """
    Method reads Role entity from the role container in directory. 
    
    required parameters:
    role.name - maps to existing role  
           
    return:
    Role   
    """    
    utils.validate_role(role)
    return roledao.read(role)

    
def find_roles(role):
    """
    Method will return a list of type Role matching all or part of Role name, Role.name.
    
    required parameters:
    role.name - maps to existing role.  May be partial name with wildcard on end - *. 
        
    return:
    Role list   
    """    
    utils.validate_role(role)
    return roledao.search(role)


def read_object(perm_obj):
    """
    Method reads permission object from perm container in directory. 
    
    required parameters:
    perm.obj_name - maps to existing perm object.   
         
    return:
    PermObj   
    """    
    utils.validate_perm_obj(perm_obj)
    return permdao.read_obj(perm_obj)

    
def find_objects(perm_obj):
    """
    Method returns a list of type PermObj that match the perm object search string. 
    
    required parameters:
    perm.obj_name - maps to existing perm object.  May be partial with wildcard on end - *.
            
    return:
    PermObj list   
    """    
    utils.validate_perm_obj(perm_obj)
    return permdao.search_objs(perm_obj)

    
def read_perm(perm):
    """
    This method returns a matching permission entity to caller. 
    
    required parameters:
    perm.obj_name - maps to already existing perm object    
    perm.op_name - maps to already existing op name
        
    optional parameters:
    perm.obj_id    
    
    return:
    Perm   
    """    
    utils.validate_perm(perm)
    return permdao.read(perm)

    
def find_perms(perm):
    """
    Method returns a list of type Permission that match the perm object search string. 
    
    required parameters:
    perm.obj_name - maps to already existing perm object.  May be partial name with wildcard on end - *.    
    perm.op_name - maps to already existing op name.  May be partial name with wildcard on end - *.
        
    optional parameters:
    perm.obj_id.  May be partial with wildcard.
        
    return:
    Perm list   
    """    
    utils.validate_perm(perm)
    return permdao.search(perm)


def assigned_users(role):
    """
    This function returns the set of users assigned to a given role. The function is valid if and only if the role is a member of the ROLES data set. 
    
    required parameters:
    role.name - maps to existing role 
        
    return:
    String list of uids   
    """    
    utils.validate_role(role)
    return roledao.get_members(role)
        
    
def assigned_roles(user):
    """
    This function returns the set of roles assigned to a given user. The function is valid if and only if the user is a member of the USERS data set.
    
    required parameters:
    user.uid - maps to existing user 
            
    return:
    Constraint list   
    """    
    utils.validate_user(user)
    usr = userdao.read(user)
    return usr.role_constraints


def perm_roles(perm):
    """
    Return a list of type String of all roles that have granted a particular permission. 
    
    required parameters:
    perm.obj_name - maps to already existing perm object    
    perm.op_name - maps to already existing op name
    
    optional parameters:
    perm.obj_id    
    
    return:
    String list of role names   
    """    
    utils.validate_perm(perm)
    out_perm = permdao.read(perm)
    return out_perm.roles


def role_perms(role):
    """
    This function returns the set of all permissions (op, obj), granted to or inherited by a given role. The function is valid if and only if the role is a member of the ROLES data set. 
    
    required parameters:
    role.name - maps to existing role  
           
    return:
    Perm list   
    """    
    utils.validate_role(role)
    return permdao.search_on_roles([role.name])


def user_perms(user):
    """
    This function returns the set of permissions a given user gets through his/her authorized roles. The function is valid if and only if the user is a member of the USERS data set. 
    
    required parameters:
    user.uid - maps to INetOrgPerson uid
         
    return:
    Perm list   
    """    
    utils.validate_user(user)
    usr = userdao.read(user)    
    return permdao.search_on_roles(usr.roles)


def perm_users(perm):
    """
    Return all users that have been granted a particular permission via their role assignments.
    
    required parameters:
    perm.obj_name - maps to already existing perm object    
    perm.op_name - maps to already existing op name
        
    optional parameters:
    perm.obj_id
        
    return:
    User list   
    """    
    utils.validate_perm(perm)
    out_perm = permdao.read(perm)
    return userdao.search_on_roles(out_perm.roles)


'''
@copyright: 2022 - Symas Corporation
'''

from rbac.model import User, Constraint

ALL = 'all'
NONE = 'none'
BEGIN_DATE = '20180101'
END_DATE = '20501010'
BEGIN_LOCK_DATE = NONE
END_LOCK_DATE = NONE
BEGIN_TIME = '0000'
END_TIME = '0000'
TIMEOUT = 15

def get_test_users(name, total):
    usrs = []
    for i in range(total):
        usrs.append(__create_test_user( name, i))
    return usrs

    
def __create_test_user(name, number):
    
    label = name + '-' + str(number)
    return User(
        uid=label, 
        description=label + ' Role',
        password='password',
        ou='py-test',                    
        sn=label,
        cn=label,                
        display_name=label,
        employee_type='test',
        title='test',
        phones=['111-111-1111', '222-222-2222'],
        mobiles=['333-333-3333', '444-444-4444'],
        emails=['foo@apache.org', 'foo2@apache.org'],
        system = False,
        props = ['name1:value1', 'name2:value2'],
        department_number = '456',
        l = 'Somewhere',
        physical_delivery_office_name = 'Someplace',
        postal_code = '55555',
        room_number = '222',                        
        constraint=Constraint(
            name=label,
            timeout=TIMEOUT,
            begin_time=BEGIN_TIME,
            end_time=END_TIME,
            begin_date=BEGIN_DATE,
            end_date=END_DATE,
            begin_lock_date=BEGIN_LOCK_DATE,
            end_lock_date=END_LOCK_DATE,
            day_mask=ALL
        )
    )


'''
@copyright: 2022 - Symas Corporation
'''

import logging
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL
import datetime
from ..util import Config

LOGGER = 'logger'
ctr = 1

def setup(ctr):
    """Function setup as many loggers as you want"""    
    now = datetime.datetime.now()
    name = Config.get(LOGGER)['file_name']
    handler = logging.FileHandler(
    #handler = logging.StreamHandler(        
        name + '-' 
        + now.strftime("%Y-%m-%d") 
        + '.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    mlog = logging.getLogger( name )
    mlog.setLevel( level_ )
    
    if Config.get(LOGGER)['console_out']:
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(formatter)
        mlog.addHandler(consoleHandler)    

    return [mlog, handler]

level_str_ = Config.get(LOGGER)['level']

if level_str_ == 'DEBUG' :
    level_ = DEBUG
elif level_str_ == 'INFO' :
    level_ = INFO
elif level_str_ == 'WARNING' :
    level_ = WARNING
elif level_str_ == 'ERROR' :
    level_ = ERROR
elif level_str_ == 'CRITICAL' :
    level_ = CRITICAL
 
is_log = False
if is_log is False:
    logger, handler_ = setup(ctr)    
    is_log = True

    
my_handlers = [handler_]
logging.basicConfig(handlers=my_handlers, level=level_)


'''
@copyright: 2022 - Symas Corporation
'''

# config attribute names
ATTRIBUTES = 'attributes'
USER_OU = 'users'
ROLE_OU = 'roles'
PERM_OU = 'perms'
SUFFIX = 'suffix'
DIT = 'dit'
      

UID = 'uid'
PROP_OC_NAME = 'ftProperties'
OU = 'ou'
INTERNAL_ID = 'ftid'
CN = 'cn'
SN = 'sn'
DN = 'dn'
CONSTRAINT = 'ftCstr'
DESC = 'description'
PROPS = 'ftProps'
CONSTRAINT = 'ftCstr'

SUCCESS = 0
OBJECT_ALREADY_EXISTS = 68
NOT_FOUND = 32
NO_SUCH_ATTRIBUTE = 16
NOT_ALLOWED_ON_NONLEAF = 66
CONFIG_BOOTSTRAP_FAILED = 126

ROLE_ALREADY_ACTIVATED_ERROR = 2001
ROLE_NOT_ACTIVATED_ERROR = 2002

USER_SEARCH_FAILED = 1000
USER_READ_FAILED = 1001
USER_ADD_FAILED = 1002
USER_UPDATE_FAILED = 1003
USER_DELETE_FAILED = 1004
USER_NOT_FOUND = 1005
USER_ID_NULL = 1006
USER_NULL = 1008
USER_PW_INVLD = 1013
USER_PW_CHK_FAILED = 1014
USER_SESS_NULL = 1030

URLE_NULL = 2003
URLE_ASSIGN_FAILED = 2004
URLE_DEASSIGN_FAILED = 2005
URLE_ACTIVATE_FAILED = 2006
URLE_DEACTIVE_FAILED = 2007
URLE_ASSIGN_EXIST = 2008
URLE_ASSIGN_NOT_EXIST = 2009
URLE_SEARCH_FAILED = 2010
URLE_ALREADY_ACTIVE = 2011
URLE_NOT_ACTIVE = 2022

ACTV_FAILED_DAY = 2050
ACTV_FAILED_DATE = 2051
ACTV_FAILED_TIME = 2052
ACTV_FAILED_TIMEOUT = 2053
ACTV_FAILED_LOCK = 2054

PERM_SEARCH_FAILED = 3000
PERM_READ_OP_FAILED = 3001
PERM_READ_OBJ_FAILED = 3002
PERM_ADD_FAILED = 3003
PERM_UPDATE_FAILED = 3004
PERM_DELETE_FAILED = 3005
PERM_NULL = 3008
PERM_OPERATION_NULL = 3009
PERM_OBJECT_NULL = 3010
PERM_OBJECT_NM_NULL = 3027
PERM_OPERATION_NM_NULL = 3026
PERM_GRANT_FAILED = 3012
PERM_REVOKE_FAILED = 3024
PERM_OP_NOT_FOUND = 3006
PERM_OBJ_NOT_FOUND = 3007
PERM_DUPLICATE = 3011
PERM_ROLE_NOT_EXIST = 3016
PERM_ROLE_SEARCH_FAILED = 3019
PERM_NOT_EXIST = 3029
PERM_OBJECT_DELETE_FAILED_NONLEAF = 3050

ROLE_SEARCH_FAILED = 5000
ROLE_READ_FAILED = 5001
ROLE_ADD_FAILED = 5002
ROLE_UPDATE_FAILED = 5003
ROLE_DELETE_FAILED = 5004
ROLE_NM_NULL = 5005
ROLE_NOT_FOUND = 5006
ROLE_NULL = 5007
ROLE_USER_ASSIGN_FAILED = 5008
ROLE_USER_DEASSIGN_FAILED = 5009
ROLE_LST_NULL = 5010
ROLE_OCCUPANT_SEARCH_FAILED = 5011
ROLE_REMOVE_OCCUPANT_FAILED = 5012

CNTR_CREATE_FAILED = 6001
CNTR_DELETE_FAILED = 6002
CNTR_NAME_NULL = 6003
CNTR_NAME_INVLD = 6004
CNTR_PARENT_NULL = 6005
CNTR_PARENT_INVLD = 6006
CNTR_NOT_FOUND = 6007
CNTR_ALREADY_EXISTS = 6008

SUFX_CREATE_FAILED = 6010
SUFX_DELETE_FAILED = 6011
SUFX_NAME_NULL = 6012
SUFX_ALREADY_EXISTS = 6016
SUFX_NOT_EXIST = 6017


'''
@copyright: 2022 - Symas Corporation
'''

import unittest
from rbac import review
from rbac.model import User, Role, Perm
from rbac.cli.utils import print_ln, print_entity


class BasicTestSuite(unittest.TestCase):
    """These tests the py-fortress review functions."""
                

class TestReivewMgr(unittest.TestCase):
    """
    Test the review funcs
    """    

    def test01_assigned_users(self):
        """
        Test the assigned users method
        """
        print_ln('test_assigned_users')
        
        try:
            rList = review.find_roles(Role(name='py-role*'))
            for rle in rList:                       
                print_ln("Assigned users role=" + rle.name)
                uList = review.assigned_users(rle)
                for user in uList:                       
                    print_ln("Assigned user=" + user, 1)
        except Exception as e:
            self.fail('test_assigned_users failed, exception=' + e.msg)


    def test02_assigned_roles(self):
        """
        Test the assigned roles method
        """
        print_ln('test_assigned_roles')
        
        try:
            uList = review.find_users(User(uid='py-user*'))
            for usr in uList:                       
                print_ln("Assigned roles user=" + usr.uid)
                rList = review.assigned_roles(usr)
                for role in rList:                       
                    print_entity(role, "Assigned role", 1)
        except Exception as e:
            self.fail('test_assigned_roles failed, exception=' + e.msg)


    def test03_perm_roles(self):
        """
        Test the perm roles method
        """
        print_ln('test16_perm_roles')
        
        try:
            pList = review.find_perms(Perm(obj_name='py-obj*', op_name='*'))
            for perm in pList:                       
                print_ln("Role Perm obj name=" + perm.obj_name + ', op=' + perm.op_name + ', id=' + perm.obj_id)
                rList = review.perm_roles(perm)
                for role in rList:
                    print_ln("Assigned role=" + role, 1)
        except Exception as e:
            self.fail('test16_perm_roles failed, exception=' + e.msg)


    def test04_role_perms(self):
        """
        Test the role perms method
        """
        print_ln('test_role_perms')
        
        try:
            rList = review.find_roles(Role(name='py-role*'))
            for rle in rList:                       
                print_ln("Perm Roles name=" + rle.name)
                pList = review.role_perms(rle)
                for perm in pList:
                    print_ln("Assigned perm obj name=" + perm.obj_name + ', op name=' + perm.op_name + ', obj id=' + perm.obj_id, 1)
        except Exception as e:
            self.fail('test_role_perms failed, exception=' + e.msg)


    def test05_user_perms(self):
        """
        Test the user perms method
        """
        print_ln('test_user_perms')
        
        try:
            uList = review.find_users(User(uid='py-user*'))
            for usr in uList:                       
                print_ln("Assigned perms user=" + usr.uid)
                pList = review.user_perms(usr)
                for perm in pList:                       
                    print_ln("Assigned perm obj name=" + perm.obj_name + ', op name=' + perm.op_name + ', obj id=' + perm.obj_id, 1)                    
        except Exception as e:
            self.fail('test_user_perms failed, exception=' + e.msg)


    def test06_perm_users(self):
        """
        Test the perm users method
        """
        print_ln('test_perm_users')
        
        try:
            pList = review.find_perms(Perm(obj_name='py-obj*', op_name='*'))
            for perm in pList:                       
                print_ln("Perm obj name=" + perm.obj_name + ', op=' + perm.op_name + ', id=' + perm.obj_id)
                uList = review.perm_users(perm)
                for user in uList:
                    print_ln("Assigned user=" + user.uid, 1)
        except Exception as e:
            self.fail('test_perm_users failed, exception=' + e.msg)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestReivewMgr('test01_assigned_users'))
    suite.addTest(TestReivewMgr('test02_assigned_roles'))
    suite.addTest(TestReivewMgr('test03_perm_roles'))
    suite.addTest(TestReivewMgr('test04_role_perms'))
    suite.addTest(TestReivewMgr('test05_user_perms'))
    suite.addTest(TestReivewMgr('test06_perm_users'))
    return suite  

 
if __name__ == '__main__':
    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())